from ..array import argpass
from .enter import EnterNode


def bind_index(node_data, parent, group, enter, update, exit, data, _):
    for i in range(len(data)):
        node = group[i] if i < len(group) else None
        if node is not None:
            node_data[node] = data[i]
            update[i] = node
        else:
            enter[i] = EnterNode(parent, data[i])

    for i in range(len(data), len(group)):
        node = group[i] if i < len(group) else None
        if node is not None:
            exit[i] = node


def bind_key(node_data, parent, group, enter, update, exit, data, key):
    node_by_key_value = {}
    key_values = [None] * len(group)
    key = argpass(key)

    for i in range(len(group)):
        node = group[i]
        if node is not None:
            key_value = key(node_data.get(node), i, group)
            if key_value in node_by_key_value:
                exit[i] = node
            else:
                node_by_key_value[key_value] = node

    for i in range(len(data)):
        key_value = key(data[i], i, data)
        node = node_by_key_value.get(key_value)
        if node is not None:
            update[i] = node
            node_data[node] = data[i]
            del node_by_key_value[key_value]
        else:
            enter[i] = EnterNode(parent, data[i])

    for i in range(len(group)):
        node = group[i]
        if node is not None and node_by_key_value.get(key_values[i]) == node:
            exit[i] = node



def contains(ring, hole):
    n = len(hole)
    for i in range(n):
        i += 1
        if c := ring_contains(ring, hole[i]):
            return c
    return 0


def ring_contains(ring, point):
    x = point[0]
    y = point[1]
    contains = -1
    n = len(ring)
    for i in range(n):
        j = i - 1 if i else n - 1
        pi = ring[i]
        xi = pi[0]
        yi = pi[1]
        pj = ring[j]
        xj = pj[0]
        yj = pj[1]
        if segment_contains(pi, pj, point):
            return 0
        if (yi > y) != (yj > y) and (x < (xj - xi) * (y - yi) / (yj - yi) + xi):
            contains = -contains
    return contains


def segment_contains(a, b, c):
    i = int(a[0] == b[0])
    return collinear(a, b, c) and within(a[i], c[i], b[i])


def collinear(a, b, c):
    return (b[0] - a[0]) * (c[1] - a[1]) == (c[0] - a[0]) * (b[1] - a[1])


def within(p, q, r):
    return p <= q <= r or r <= q <= p



from ..number import number
from .parse import parseCss, parseSvg


def interpolateTransform(parse, pxComma, pxParen, degParen):
    def pop(s):
        return s.pop() + " " if s else ""

    def translate(xa, ya, xb, yb, s, q):
        if xa != xb or ya != yb:
            s.extend(["translate(", None, pxComma, None, pxParen])
            q.extend(
                [
                    {"i": len(s) - 4, "x": number(xa, xb)},
                    {"i": len(s) - 2, "x": number(ya, yb)},
                ]
            )
        elif xb or yb:
            s.append(f"translate({xb}{pxComma}{yb}{pxParen}")

    def rotate(a, b, s, q):
        if a != b:
            if a - b > 180:
                b += 360
            elif b - a > 180:
                a += 360
            q.append({"i": len(s), "x": number(a, b)})
            s.extend([pop(s) + "rotate(", None, degParen])
        elif b:
            s.append(f"{pop(s)}rotate({b}{degParen}")

    def skewX(a, b, s, q):
        if a != b:
            q.append({"i": len(s), "x": number(a, b)})
            s.extend([pop(s) + "skewX(", None, degParen])
        elif b:
            s.append(f"{pop(s)}skewX({b}{degParen}")

    def scale(xa, ya, xb, yb, s, q):
        if xa != xb or ya != yb:
            s.extend([pop(s) + "scale(", None, ",", None, ")"])
            q.extend(
                [
                    {"i": len(s) - 4, "x": number(xa, xb)},
                    {"i": len(s) - 2, "x": number(ya, yb)},
                ]
            )
        elif xb != 1 or yb != 1:
            s.append(f"{pop(s)}scale({xb},{yb})")

    def interpolator(a, b):
        s = []
        q = []
        a = parse(a)
        b = parse(b)
        translate(
            a["translateX"], a["translateY"], b["translateX"], b["translateY"], s, q
        )
        rotate(a["rotate"], b["rotate"], s, q)
        skewX(a["skewX"], b["skewX"], s, q)
        scale(a["scaleX"], a["scaleY"], b["scaleX"], b["scaleY"], s, q)

        def interpolate(t):
            for o in q:
                s[o["i"]] = o["x"](t)
            return "".join(s)

        return interpolate

    return interpolator


interpolateTransformCss = interpolateTransform(parseCss, "px, ", "px)", "deg)")
interpolateTransformSvg = interpolateTransform(parseSvg, ", ", ")", ")")



from .freedman_diaconis import threshold_freedman_diaconis
from .scott import threshold_scott
from .sturges import threshold_sturges

__all__ = [
    "threshold_freedman_diaconis",
    "threshold_scott",
    "threshold_sturges",
]



from functools import cmp_to_key

from ..types import Point2D
from .cross import cross


def lexicographic_order(a: Point2D, b: Point2D) -> float:
    return a[0] - b[0] or a[1] - b[1]


def compute_upper_hull_indexes(points: list[Point2D]) -> list[int]:
    indexes = [0, 1] + [None] * (len(points) - 2)
    size = 2

    for i in range(2, len(points)):
        while (
            size > 1
            and cross(points[indexes[size - 2]], points[indexes[size - 1]], points[i])
            <= 0
        ):
            size -= 1
        indexes[size] = i
        size += 1

    return indexes[:size]


def hull(points: list[Point2D]) -> list[Point2D] | None:
    """
    Returns the convex hull of the specified points using `Andrew's monotone
    chain algorithm
    <https://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/Convex_hull/Monotone_chain>`_.
    The returned hull is represented as an array containing a subset of the
    input points arranged in counterclockwise order. Returns :code:`None` if
    points has fewer than three elements.

    Parameters
    ----------
    points : list[Point2D]
        List of points

    Returns
    -------
    list[Point2D] | None
        Convex hull or :code:`None` if :code:`points` has fewer than three
        elements.

    Examples
    --------
    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> d3.polygon_hull(points)
    [[38, 2], [2, 2], [6, 8], [10, 10], [12, 10], [24, 8], [35, 5]]
    """
    n = len(points)
    if n < 3:
        return None

    sorted_points = [None] * n
    flipped_points = [None] * n

    for i, point in enumerate(points):
        sorted_points[i] = [point[0], point[1], i]

    sorted_points.sort(key=cmp_to_key(lexicographic_order))

    for i, point in enumerate(sorted_points):
        flipped_points[i] = [point[0], -point[1]]

    upper_indexes = compute_upper_hull_indexes(sorted_points)
    lower_indexes = compute_upper_hull_indexes(flipped_points)

    skip_left = lower_indexes[0] == upper_indexes[0]
    skip_right = lower_indexes[-1] == upper_indexes[-1]
    hull = []

    for i in range(len(upper_indexes) - 1, -1, -1):
        hull.append(points[sorted_points[upper_indexes[i]][2]])

    for i in range(skip_left, len(lower_indexes) - skip_right):
        hull.append(points[sorted_points[lower_indexes[i]][2]])

    return hull



from .format_decimal import format_decimal_parts


def format_rounded(x, p):
    d = format_decimal_parts(x, p)
    if not d:
        return str(x)
    coefficient, exponent = d
    if exponent < 0:
        return "0." + "0" * (-1 - exponent) + coefficient
    elif len(coefficient) > exponent + 1:
        return coefficient[: exponent + 1] + "." + coefficient[exponent + 1 :]
    else:
        return coefficient + "0" * (exponent - len(coefficient) + 1)



from collections.abc import Callable
from math import degrees, pi, radians

from ...types import Point2D
from ..common import Projection, RawProjection
from .projection import ProjectionMutator


class ConicProjection(ProjectionMutator):
    def __init__(
        self, project_func: Callable[..., RawProjection], phi0: float, phi1: float
    ):
        ProjectionMutator.__init__(self, project_func(phi0, phi1))
        self._project_func = project_func
        self._phi0 = phi0
        self._phi1 = phi1

    def parallels(self, point: Point2D | None = None) -> Projection:
        """
        The two standard parallels that define the map layout in conic
        projections. If :code:`point` is :code:`None`, returns the current
        parallels parameters.

        Parameters
        ----------
        point : Point2D | None
            2D Point

        Returns
        -------
        Projection
            Itself
        """
        if point is None:
            return [degrees(self._phi0), degrees(self._phi1)]
        else:
            return self.set_project(
                self._project_func(radians(point[0]), radians(point[1]))
            )


def conic_projection(project: Callable[..., RawProjection]) -> ConicProjection:
    return ConicProjection(project, 0, pi / 3)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "e5f5e0a1d99b31a354",
    "edf8e9bae4b374c476238b45",
    "edf8e9bae4b374c47631a354006d2c",
    "edf8e9c7e9c0a1d99b74c47631a354006d2c",
    "edf8e9c7e9c0a1d99b74c47641ab5d238b45005a32",
    "f7fcf5e5f5e0c7e9c0a1d99b74c47641ab5d238b45005a32",
    "f7fcf5e5f5e0c7e9c0a1d99b74c47641ab5d238b45006d2c00441b",
]
SCHEME_GREENS = list(map(colors, scheme))
interpolate_greens = ramp(SCHEME_GREENS)



from __future__ import annotations

from collections.abc import Callable
from math import cos, degrees, radians, sin, sqrt
from typing import Any

from ...array import argpass
from ...types import GeoJSON, Point2D, Vec2D
from ..clip import geo_clip_antimeridian, geo_clip_circle, geo_clip_rectangle
from ..common import PolygonStream, Projection, RawProjection, SpatialTransform
from ..compose import Compose
from ..rotation import RotateRadians
from ..transform import GeoTransformer
from .fit import fit_extent, fit_height, fit_size, fit_width
from .resample import resample


def point(self, x: float, y: float):
    self._stream.point(radians(x), radians(y))


transform_radians = GeoTransformer({"point": point})


def transform_rotate(rotate: RotateRadians) -> GeoTransformer:
    def point(self, x: float, y: float):
        a, b = rotate(x, y)
        return self._stream.point(a, b)

    return GeoTransformer({"point": point})


class ScaleTranslate:
    def __init__(self, k: float, dx: float, dy: float, sx: float, sy: float):
        self._k = k
        self._dx = dx
        self._dy = dy
        self._sx = sx
        self._sy = sy

    def __call__(self, x: float, y: float) -> Point2D:
        x *= self._sx
        y *= self._sy
        return [self._dx + self._k * x, self._dy - self._k * y]

    def invert(self, x: float, y: float) -> Point2D:
        return [
            (x - self._dx) / self._k * self._sx,
            (self._dy - y) / self._k * self._sy,
        ]


class ScaleTranslateRotate:
    def __init__(
        self, k: float, dx: float, dy: float, sx: float, sy: float, alpha: float
    ):
        self._cos_alpha = cos(alpha)
        self._sin_alpha = sin(alpha)
        self._sx = sx
        self._sy = sy
        self._k = k
        self._dx = dx
        self._dy = dy
        self._a = self._cos_alpha * self._k
        self._b = self._sin_alpha * self._k
        self._ai = self._cos_alpha / self._k
        self._bi = self._sin_alpha / self._k
        self._ci = (self._sin_alpha * self._dy - self._cos_alpha * self._dx) / self._k
        self._fi = (self._sin_alpha * self._dx + self._cos_alpha * self._dy) / self._k

    def __call__(self, x: float, y: float) -> Point2D:
        x *= self._sx
        y *= self._sy
        return [
            self._a * x - self._b * y + self._dx,
            self._dy - self._b * x - self._a * y,
        ]

    def invert(self, x: float, y: float) -> Point2D:
        return [
            self._sx * (self._ai * x - self._bi * y + self._ci),
            self._sy * (self._fi - self._bi * x - self._ai * y),
        ]


def scale_translate_rotate(
    k: float, dx: float, dy: float, sx: float, sy: float, alpha: float
) -> SpatialTransform:
    if alpha:
        return ScaleTranslateRotate(k, dx, dy, sx, sy, alpha)
    else:
        return ScaleTranslate(k, dx, dy, sx, sy)


def identity(x):
    return x


class RawProjectionIdentity:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        return [lambda_, phi]

    def invert(self, x: float, y: float) -> Point2D:
        return self(x, y)


class ProjectionMutator(Projection):
    """
    Projections transform spherical polygonal geometry to planar polygonal
    geometry. Several classes of standard projections are provided:

    - Azimuthal projections
    - Conic projections
    - Cylindrical projections

    Parameters
    ----------
    project : RawProjection
        Projection object
    """

    def __init__(self, project: RawProjection):
        self._project = project
        self._invert = (
            self._invert_default if hasattr(project, "invert") else self._invert_error
        )
        self._k = 150
        self._x = 480
        self._y = 250
        self._lambda = 0
        self._phi = 0
        self._delta_lambda = 0
        self._delta_phi = 0
        self._delta_gamma = 0
        self._rotate = None
        self._alpha = 0
        self._sx = 1
        self._sy = 1
        self._theta = None
        self._preclip = geo_clip_antimeridian
        self._x0 = None
        self._y0 = None
        self._x1 = None
        self._y1 = None
        self._postclip = identity
        self._delta2 = 0.5
        self._project_resample = None
        self._project_transform = None
        self._project_rotate_transform = None
        self._cache = None
        self._cache_stream = None
        self.recenter()

    def __call__(self, point: Point2D) -> Point2D:
        """
        Returns a new array :code:`[x, y]` (typically in pixels) representing
        the projected point of the given point. The point must be specified as
        a two-element array :code:`[longitude, latitude]` in degrees. May
        return :code:`None` if the specified point has no defined projected position,
        such as when the point is outside the clipping bounds of the
        projection.

        Parameters
        ----------
        point : Point2D
            Point :code:`[longitude, latitude]`

        Returns
        -------
        Point2D
            New projected point :code:`[x, y]`
        """
        return self._project_rotate_transform(radians(point[0]), radians(point[1]))

    def invert(self, point: Point2D) -> Point2D:
        """
        Returns a new array :code:`[longitude, latitude]` in degrees
        representing the unprojected point of the given projected point. The
        point must be specified as a two-element array :code:`[x, y]`
        (typically in pixels). May return null if the specified point has no
        defined projected position, such as when the point is outside the
        clipping bounds of the projection.

        This method is only defined on invertible projections.

        Parameters
        ----------
        point : Point2D
            2D point where coordinates are in pixels

        Returns
        -------
        Point2D
            2D point where coordinates are in degrees
        """
        return self._invert(point)

    def _invert_default(self, point: Point2D) -> Point2D:
        point = self._project_rotate_transform.invert(point[0], point[1])
        if point is not None:
            return [degrees(point[0]), degrees(point[1])]

    def _invert_error(self, point: Point2D):
        raise NotImplementedError("Projection does not support invert method.")

    def stream(self, stream: PolygonStream) -> PolygonStream:
        """
        Returns a projection stream for the specified output stream. Any input
        geometry is projected before being streamed to the output stream. A
        typical projection involves several geometry transformations: the input
        geometry is first converted to radians, rotated on three axes, clipped
        to the small circle or cut along the antimeridian, and lastly projected
        to the plane with adaptive resampling, scale and translation.

        Parameters
        ----------
        stream : PolygonStream
            Stream object

        Returns
        -------
        PolygonStream
            Modified stream object
        """
        if self._cache and self._cache_stream == stream:
            return self._cache
        self._cache_stream = stream
        clipped_stream = self._preclip(self._project_resample(self._postclip(stream)))
        self._cache = transform_radians(transform_rotate(self._rotate)(clipped_stream))
        return self._cache

    def set_project(self, project: RawProjection) -> ProjectionMutator:
        """
        Updates the raw projection object and returns itself

        Parameters
        ----------
        project : RawProjection
            Raw projection object

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._project = project
        self._invert = (
            self._invert_default if hasattr(project, "invert") else self._invert_error
        )
        return self.recenter()

    def set_preclip(
        self, preclip: Callable[[PolygonStream], PolygonStream]
    ) -> ProjectionMutator:
        """
        If preclip is specified, sets the projection's spherical clipping to
        the specified function and returns the projection; preclip is a
        function that takes a projection stream and returns a clipped stream.

        Parameters
        ----------
        preclip : Callable[[PolygonStream], PolygonStream]
            Preclip function

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._preclip = preclip
        self._theta = None
        return self.reset()

    def set_postclip(
        self, postclip: Callable[[PolygonStream], PolygonStream]
    ) -> ProjectionMutator:
        """
        If postclip is specified, sets the projection's Cartesian clipping to
        the specified function and returns the projection; postclip is a
        function that takes a projection stream and returns a clipped stream.

        Parameters
        ----------
        postclip : Callable[[PolygonStream], PolygonStream]
            Postclip function

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._postclip = postclip
        self._x0 = None
        self._x1 = None
        self._y0 = None
        self._y1 = None
        return self.reset()

    def set_clip_angle(self, clip_angle: float | None = None) -> ProjectionMutator:
        """
        If angle is specified, sets the projection's clipping circle radius to
        the specified angle in degrees and returns the projection. If angle is
        :code:`None`, switches to antimeridian cutting rather than small-circle
        clipping. If angle is not specified, returns the current clip angle
        which defaults to :code:`None`. Small-circle clipping is independent of
        viewport clipping via projection.clipExtent.

        Parameters
        ----------
        clip_angle : float | None
            Clip angle

        Returns
        -------
        ProjectionMutator
            Itself
        """
        if clip_angle:
            self._theta = radians(clip_angle)
            self._preclip = geo_clip_circle(self._theta)
        else:
            self._theta = None
            self._preclip = geo_clip_antimeridian
        return self.reset()

    def set_clip_extent(
        self, clip_extent: tuple[Point2D, Point2D] | None = None
    ) -> ProjectionMutator:
        """
        If extent is specified, sets the projection's viewport clip extent to
        the specified bounds in pixels and returns the projection. The extent
        bounds are specified as an array :math:`[[x_0, y_0], [x_1, y_1]]`,
        where :math:`x_0` is the left-side of the viewport, :math:`y_0` is the
        top, :math:`x_1` is the right and :math:`y_1` is the bottom. If extent
        is :code:`None`, no viewport clipping is performed. If extent is not
        specified, returns the current viewport clip extent which defaults to
        :code:`None`. Viewport clipping is independent of small-circle clipping
        via projection.clipAngle.

        Parameters
        ----------
        clip_extent : tuple[Point2D, Point2D] | None


        Returns
        -------
        ProjectionMutator
            Itself
        """
        if clip_extent is None:
            self._x0 = None
            self._x1 = None
            self._y0 = None
            self._y1 = None
            self._postclip = identity
        else:
            self._x0 = clip_extent[0][0]
            self._x1 = clip_extent[1][0]
            self._y0 = clip_extent[0][1]
            self._y1 = clip_extent[1][1]
            self._postclip = geo_clip_rectangle(self._x0, self._y0, self._x1, self._y1)
        return self.reset()

    def scale(self, k: float) -> ProjectionMutator:
        """
        If scale is specified, sets the projection's scale factor to the
        specified value and returns the projection. The scale factor
        corresponds linearly to the distance between projected points; however,
        absolute scale factors are not equivalent across projections.

        Parameters
        ----------
        k : float
            Scale factor

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._k = k
        return self.recenter()

    def translate(self, translation: Vec2D) -> ProjectionMutator:
        """
        If translate is specified, sets the projection's translation offset to
        the specified two-element array :code:`[tx, ty]` and returns the
        projection. The translation offset determines the pixel coordinates of
        the projection's center. The default translation offset places
        :math:`[0°,0°]` at the center of a :math:`960 \\times 500` area.

        Parameters
        ----------
        translation : Vec2D
            Translation 2D vector

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._x = translation[0]
        self._y = translation[1]
        return self.recenter()

    def set_center(self, center: Point2D) -> ProjectionMutator:
        """
        If center is specified, sets the projection's center to the specified
        center, a two-element array of :code:`[longitude, latitude]` in degrees
        and returns the projection.

        Parameters
        ----------
        center : Point2D
            Center point

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._lambda = radians(center[0])
        self._phi = radians(center[1])
        return self.recenter()

    def rotate(
        self, angles: tuple[float, float] | tuple[float, float, float]
    ) -> ProjectionMutator:
        """
        If angles is specified, sets the projection's three-axis spherical
        rotation to the specified value, which must be a two- or three-element
        array of numbers :math:`[\\lambda, \\phi, \\gamma]` specifying the
        rotation angles in degrees about each spherical axis. (These correspond
        to yaw, pitch and roll).

        Parameters
        ----------
        angles : tuple[float, float] | tuple[float, float, float]
            :math:`[\\lambda, \\phi, \\gamma]` values or :math:`[\\lambda, \\phi]` values

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._delta_lambda = radians(angles[0])
        self._delta_phi = radians(angles[1])
        self._delta_gamma = radians(angles[2]) if len(angles) > 2 else 0
        return self.recenter()

    def set_angle(self, angle: float) -> ProjectionMutator:
        """
        If angle is specified, sets the projection's post-projection planar
        rotation angle to the specified angle in degrees and returns the
        projection.

        Parameters
        ----------
        angle : float
            Angle value

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._alpha = radians(angle)
        return self.recenter()

    def set_reflect_x(self, reflect_x: bool) -> ProjectionMutator:
        """
        If reflect is specified, sets whether or not the x-dimension is
        reflected (negated) in the output. This can be useful to display sky
        and astronomical data with the orb seen from below: right ascension
        (eastern direction) will point to the left when North is pointing up.

        Parameters
        ----------
        reflect_x : bool
            Boolean value

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._sx = -1 if reflect_x else 1
        return self.recenter()

    def set_reflect_y(self, reflect_y: bool) -> ProjectionMutator:
        """
        If reflect is specified, sets whether or not the y-dimension is
        reflected (negated) in the output. This is especially useful for
        transforming from standard spatial reference systems, which treat
        positive y as pointing up, to display coordinate systems such as Canvas
        and SVG, which treat positive y as pointing down.

        Parameters
        ----------
        reflect_y : bool
            Boolean value

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._sy = -1 if reflect_y else 1
        return self.recenter()

    def set_precision(self, precision: float) -> ProjectionMutator:
        """
        If precision is specified, sets the threshold for the projection's
        adaptive resampling to the specified value in pixels and returns the
        projection. This value corresponds to the Douglas–Peucker distance.

        Parameters
        ----------
        precision : float
            Precision value

        Returns
        -------
        ProjectionMutator
            Itself
        """
        self._delta2 = precision * precision
        self._project_resample = resample(self._project_transform, self._delta2)
        return self.reset()

    def fit_extent(
        self, extent: tuple[Point2D, Point2D], obj: GeoJSON
    ) -> ProjectionMutator:
        """
        Sets the projection's scale and translate to fit the specified GeoJSON
        object in the center of the given extent. The extent is specified as an
        array :math:`[[x_0, y_0], [x_1, y_1]]`, where :math:`x_0` is the left
        side of the bounding box, :math:`y_0` is the top, :math:`x_1`₁ is the
        right and :math:`y_1` is the bottom. Returns the projection.

        Any clip extent is ignored when determining the new scale and
        translate. The precision used to compute the bounding box of the given
        object is computed at an effective scale of 150.

        Parameters
        ----------
        extent : tuple[Point2D, Point2D]
            Extent values
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        ProjectionMutator
            Itself
        """
        return fit_extent(self, extent, obj)

    def fit_size(self, size: Point2D, obj: GeoJSON) -> ProjectionMutator:
        """
        A convenience method for projection.fit_extent where the top-left
        corner of the extent is [0, 0].

        Parameters
        ----------
        size : Point2D
            Size values
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        ProjectionMutator
            Itself
        """
        return fit_size(self, size, obj)

    def fit_width(self, width: float, obj: GeoJSON) -> ProjectionMutator:
        """
        A convenience method for projection.fit_size where the height is
        automatically chosen from the aspect ratio of object and the given
        constraint on width.

        Parameters
        ----------
        width : float
            Width value
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        ProjectionMutator
            Itself
        """
        return fit_width(self, width, obj)

    def fit_height(self, height: float, obj: GeoJSON) -> ProjectionMutator:
        """
        A convenience method for projection.fit_size where the width is
        automatically chosen from the aspect ratio of object and the given
        constraint on height.

        Parameters
        ----------
        height : float
            Height value
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        ProjectionMutator
            Itself
        """
        return fit_height(self, height, obj)

    def recenter(self) -> ProjectionMutator:
        center = scale_translate_rotate(self._k, 0, 0, self._sx, self._sy, self._alpha)(
            *self._project(self._lambda, self._phi)
        )
        transform = scale_translate_rotate(
            self._k,
            self._x - center[0],
            self._y - center[1],
            self._sx,
            self._sy,
            self._alpha,
        )
        self._rotate = RotateRadians(
            self._delta_lambda, self._delta_phi, self._delta_gamma
        )
        self._project_transform = Compose(self._project, transform)
        self._project_rotate_transform = Compose(self._rotate, self._project_transform)
        self._project_resample = resample(self._project_transform, self._delta2)
        return self.reset()

    def reset(self) -> ProjectionMutator:
        self._cache = None
        self._cache_stream = None
        return self

    def get_preclip(self) -> Callable[[PolygonStream], PolygonStream]:
        return self._preclip

    def get_postclip(self) -> Callable[[PolygonStream], PolygonStream]:
        return self._postclip

    def get_clip_angle(self) -> float:
        return radians(self._theta)

    def get_clip_extent(self) -> tuple[Point2D, Point2D] | None:
        if self._x0 is None:
            return None
        else:
            return [[self._x0, self._y0], [self._x1, self._y1]]

    def get_scale(self) -> float:
        return self._k

    def get_translation(self) -> Vec2D:
        return [self._x, self._y]

    def get_center(self) -> Point2D:
        return [degrees(self._lambda), degrees(self._phi)]

    def get_rotation(self) -> tuple[float, float, float]:
        return [
            degrees(self._delta_lambda),
            degrees(self._delta_phi),
            degrees(self._delta_gamma),
        ]

    def get_angle(self) -> float:
        return degrees(self._alpha)

    def get_reflect_x(self) -> bool:
        return self._sx < 0

    def get_reflect_y(self) -> bool:
        return self._sy < 0

    def get_precision(self) -> float:
        return sqrt(self._delta2)


def geo_projection(project: RawProjection) -> Projection:
    """
    Constructs a new projection from the specified raw projection :code:`project`.
    The :code:`project` function takes the longitude and latitude of a given
    point in radians, often referred to as :code:`lambda` (:math:`\\lambda`)
    and :code:`phi` (:math:`\\phi`), and returns a two-element array :code:`[x,
    y]` representing its unit projection. The project function does not need to
    scale or translate the point, as these are applied automatically by
    projection.scale, projection.translate, and projection.center. Likewise,
    the project function does not need to perform any spherical rotation, as
    projection.rotate is applied prior to projection.

    Parameters
    ----------
    project : RawProjection
        Projection object

    Returns
    -------
    Projection
        Projection object
    """
    return ProjectionMutator(project)


def geo_projection_mutator(
    project: Callable[..., RawProjection],
) -> Callable[..., Projection]:
    """
    Constructs a new projection from the specified raw projection factory and
    returns a mutate function to call whenever the raw projection changes. The
    factory must return a raw projection. The returned mutate function returns
    the wrapped projection. For example, a conic projection typically has two
    configurable parallels.

    Parameters
    ----------
    project : Callable[..., RawProjection]
        Function which generates a projection object

    Returns
    -------
    Callable[..., Projection]
        Projection object
    """
    projection = ProjectionMutator(RawProjectionIdentity())

    def projection_constructor(*args: Any) -> Projection:
        return projection.set_project(argpass(project)(*args))

    return projection_constructor



from collections.abc import Callable
from math import floor


def blur(values: list[float], r: float) -> list[float]:
    """
    Blurs an array of data in-place by applying three iterations of a moving
    average transform (box filter) for a fast approximation of a Gaussian
    kernel of the given radius, a non-negative number. Returns the given data.

    Parameters
    ----------
    values : list[float]
       Data
    r : float
        Radius

    Returns
    -------
    list[float]
        Blurred data

    Examples
    --------
    >>> d3.blur([0, 0, 1, 0, 0], 1)
    [0.14814814814814814, 0.2222222222222222, 0.25925925925925924, 0.22222222222222224, 0.1481481481481482]
    """
    if r < 0:
        raise ValueError("r cannot be negative")
    length = len(values)
    if not length or not r:
        return values
    blur = blurf(r)
    tmp = values[:]
    blur(values, tmp, 0, length, 1)
    blur(tmp, values, 0, length, 1)
    blur(values, tmp, 0, length, 1)
    return values


def blurf(
    radius: float,
) -> Callable[[list[float], list[float], int, int, int], None]:
    radius0 = floor(radius)
    if radius0 == radius:
        return bluri(radius)
    t = radius - radius0
    w = 2 * radius + 1

    def local_blur(T, S, start, stop, step):
        stop -= step
        if stop < start:
            return
        sum = radius0 * S[start]
        s0 = step * radius0
        s1 = s0 + step
        for i in range(start, start + s0, step):
            sum += S[min(stop, i)]
        for i in range(start, stop + step, step):
            sum += S[min(stop, i + s0)]
            T[i] = (sum + t * (S[max(start, i - s1)] + S[min(stop, i + s1)])) / w
            sum -= S[max(start, i - s0)]

    return local_blur


def blur2d(
    blur: Callable[[float], Callable[[list[float], list[float], int, int, int], None]],
) -> Callable[[dict, float, float | None], dict]:
    def local_blur2(data: dict, rx: float, ry: float | None = None):
        ry = rx if ry is None else ry
        if rx < 0.0:
            raise ValueError("rx cannot be negative")
        if ry < 0.0:
            raise ValueError("ry cannot be negative")
        values = data["data"]
        width = floor(data["width"])
        height = floor(data.get("height", len(values) / width))
        if width < 0:
            raise ValueError("Invalid width")
        if height < 0:
            raise ValueError("Invalid height")
        if not width or not height or (not rx and not ry):
            return data
        blurx = blur(rx) if rx else None
        blury = blur(ry) if ry else None
        tmp = values[:]
        if blurx and blury:
            blurh(blurx, tmp, values, width, height)
            blurh(blurx, values, tmp, width, height)
            blurh(blurx, tmp, values, width, height)
            blurv(blury, values, tmp, width, height)
            blurv(blury, tmp, values, width, height)
            blurv(blury, values, tmp, width, height)
        elif blurx:
            blurh(blurx, values, tmp, width, height)
            blurh(blurx, tmp, values, width, height)
            blurh(blurx, values, tmp, width, height)
        else:
            blurv(blury, values, tmp, width, height)
            blurv(blury, tmp, values, width, height)
            blurv(blury, values, tmp, width, height)
        return {"data": values, "width": width, "height": height}

    return local_blur2


def blurh(
    blur: Callable[[list[float], list[float], int, int, int], None],
    T: list[float],
    S: list[float],
    w: float,
    h: float,
):
    for y in range(0, w * h, w):
        blur(T, S, y, y + w, 1)


def blurv(
    blur: Callable[[list[float], list[float], int, int, int], None],
    T: list[float],
    S: list[float],
    w: float,
    h: float,
):
    n = w * h
    for x in range(w):
        blur(T, S, x, x + n, w)


def blurf_image(
    radius: float,
) -> Callable[[list[float], list[float], int, int, int], None]:
    blur = blurf(radius)

    def local_blur(
        T: list[float],
        S: list[float],
        start: int,
        stop: int,
        step: int,
    ):
        start <<= 2
        stop <<= 2
        step <<= 2
        blur(T, S, start, stop, step)
        blur(T, S, start + 1, stop + 1, step)
        blur(T, S, start + 2, stop + 2, step)
        blur(T, S, start + 3, stop + 3, step)

    return local_blur


def bluri(
    radius: float,
) -> Callable[[list[float], list[float], int, int, int], None]:
    w = 2 * radius + 1

    def local_blur(
        T: list[float],
        S: list[float],
        start: int,
        stop: int,
        step: int,
    ):
        stop -= step
        if stop < start:
            return
        sum = radius * S[start]
        s = step * radius
        for i in range(start, int(start + s), step):
            sum += S[min(stop, i)]
        for i in range(start, stop + step, step):
            sum += S[min(stop, int(i + s))]
            T[i] = sum / w
            sum -= S[max(start, int(i - s))]

    return local_blur


blur2 = blur2d(blurf)
blur2.__doc__ = """
Blurs a matrix of the given width and height in-place by applying a horizontal
blur of radius rx and a vertical blur of radius ry (which defaults to rx). The
matrix values data are stored in a flat (one-dimensional) array.

Parameters
----------
data : dict
    Dictionary with three keys: :code:`"data"`, :code:`"width"` and
    :code:`"height"`. If :code:`"height"` is not specified, it is inferred from
    the given width and length of data.
rx : float
    Radius x
ry : float | None
    Radius y

Returns
-------
dict
    Returns the blurred matrix

Examples
--------
>>> data = [
... 0, 0, 0, 0, 0,
... 0, 0, 1, 0, 0,
... 0, 1, 1, 1, 0,
... 0, 0, 1, 0, 0,
... 0, 0, 0, 0, 0,
... ]
>>> d3.blur2({"data": data, "width": 5, "height": 5}, 1)
{'data': [0.1316872427983539, 0.1755829903978052, 0.20027434842249658, 0.17558299039780523, 0.13168724279835398, 0.1755829903978052, 0.23045267489711932, 0.262002743484225, 0.23045267489711938, 0.1755829903978053, 0.20027434842249656, 0.262002743484225, 0.29766803840877915, 0.262002743484225, 0.20027434842249667, 0.1755829903978052, 0.23045267489711932, 0.262002743484225, 0.23045267489711938, 0.1755829903978053, 0.1316872427983539, 0.1755829903978052, 0.20027434842249656, 0.17558299039780523, 0.13168724279835398], 'width': 5, 'height': 5}
"""
blur_image = blur2d(blurf_image)
blur_image.__doc__ = """
Blurs the given image in-place, blurring each of the RGBA layers
independently by applying an horizontal blur of radius rx and a vertical blur
of radius ry (which defaults to rx).

Parameters
----------
data : dict
    Dictionary with three keys: :code:`"data"`, :code:`"width"` and
    :code:`"height"`. If :code:`"height"` is not specified, it is inferred from
    the given width and length of data.
rx : float
    Radius x
ry : float | None
    Radius y

Returns
-------
dict
    Returns the blurred image
"""



from math import fsum, nan, sqrt

from ...array import argpass
from ..common import PolygonStream


@argpass
def noop():
    return


class LengthStream(PolygonStream):
    def __init__(self):
        self._point = noop
        self._length_sum = []
        self._length_ring = False
        self._x00 = nan
        self._y00 = nan
        self._x0 = nan
        self._y0 = nan

    def line_start(self):
        self._point = self._length_point_first

    def line_end(self):
        if self._length_ring:
            self._length_point(self._x00, self._y00)
        self._point = noop

    def polygon_start(self):
        self._length_ring = True

    def polygon_end(self):
        self._length_ring = False

    def point(self, x: float, y: float):
        return self._point(x, y)

    def result(self):
        length = fsum(self._length_sum)
        self._length_sum = []
        return length

    def _length_point_first(self, x: float, y: float):
        self._point = self._length_point
        self._x00 = self._x0 = x
        self._y00 = self._y0 = y

    def _length_point(self, x: float, y: float):
        self._x0 -= x
        self._y0 -= y
        self._length_sum.append(sqrt(self._x0 * self._x0 + self._y0 * self._y0))
        self._x0 = x
        self._y0 = y



from __future__ import annotations

from math import asin, pi, sin

from .tqmt import tqmt

TAU = 2 * pi
AMPLITUDE = 1
PERIOD = 0.3


class ElasticIn:
    """
    Elastic easing, like a rubber band.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_elastic_in(0.2)
    -0.0014662756598240474
    >>> d3.ease_elastic_in.amplitude(2)(0.2)
    -0.005865102639296188
    >>> d3.ease_elastic_in.period(0.5)(0.2)
    -0.0023724838544719874
    >>> d3.ease_elastic_in.amplitude(2).period(0.5)(0.2)
    -0.00535803787473666
    """

    def __init__(self, amplitude: float, period: float):
        self._amplitude = max(1, amplitude)
        self._period = period / TAU
        self._s = asin(1 / self._amplitude) * self._period

    def __call__(self, t: float) -> float:
        """
        Elastic easing, like a rubber band.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_elastic_in(0.2)
        -0.0014662756598240474
        >>> d3.ease_elastic_in.amplitude(2)(0.2)
        -0.005865102639296188
        >>> d3.ease_elastic_in.period(0.5)(0.2)
        -0.0023724838544719874
        >>> d3.ease_elastic_in.amplitude(2).period(0.5)(0.2)
        -0.00535803787473666
        """
        t -= 1
        return self._amplitude * tqmt(-t) * sin((self._s - t) / self._period)

    def amplitude(self, amplitude: float) -> ElasticIn:
        """
        Returns a new easing function with :code:`amplitude` value as
        parameter.

        Parameters
        ----------
        amplitude : float
            Amplitude value

        Returns
        -------
        ElasticIn
            New easing function
        """
        return ElasticIn(amplitude, self._period * TAU)

    def period(self, period: float) -> ElasticIn:
        """
        Returns a new easing function with :code:`period` value as parameter.

        Parameters
        ----------
        period : float
            Amplitude value

        Returns
        -------
        ElasticIn
            New easing function
        """
        return ElasticIn(self._amplitude, period)


ease_elastic_in = ElasticIn(AMPLITUDE, PERIOD)


class ElasticOut:
    """
    Reverse elastic easing; equivalent to :code:`1 - d3.elastic_in(1 - t)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_elastic_out(0.2)
    1.1246334310850439
    >>> d3.ease_elastic_out.amplitude(2)(0.2)
    1.498533724340176
    >>> d3.ease_elastic_out.period(0.5)(0.2)
    1.2016611276301188
    >>> d3.ease_elastic_out.amplitude(2).period(0.5)(0.2)
    0.9478890359076215
    """

    def __init__(self, amplitude: float, period: float):
        self._amplitude = max(1, amplitude)
        self._period = period / TAU
        self._s = asin(1 / self._amplitude) * self._period

    def __call__(self, t: float) -> float:
        """
        Reverse elastic easing; equivalent to :code:`1 - d3.elastic_in(1 - t)`.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_elastic_out(0.2)
        1.1246334310850439
        >>> d3.ease_elastic_out.amplitude(2)(0.2)
        1.498533724340176
        >>> d3.ease_elastic_out.period(0.5)(0.2)
        1.2016611276301188
        >>> d3.ease_elastic_out.amplitude(2).period(0.5)(0.2)
        0.9478890359076215
        """
        return 1 - self._amplitude * tqmt(t) * sin((t + self._s) / self._period)

    def amplitude(self, amplitude: float) -> ElasticOut:
        """
        Returns a new easing function with :code:`amplitude` value as
        parameter.

        Parameters
        ----------
        amplitude : float
            Amplitude value

        Returns
        -------
        ElasticOut
            New easing function
        """
        return ElasticOut(amplitude, self._period * TAU)

    def period(self, period: float) -> ElasticOut:
        """
        Returns a new easing function with :code:`period` value as parameter.

        Parameters
        ----------
        period : float
            Amplitude value

        Returns
        -------
        ElasticOut
            New easing function
        """
        return ElasticOut(self._amplitude, period)


ease_elastic_out = ElasticOut(AMPLITUDE, PERIOD)


class ElasticInOut:
    """
    Symmetric elastic easing; scales :func:`d3.ease_elastic_in
    <detroit.ease_elastic_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_elastic_out <detroit.ease_elastic_out>` :math:`\\forall t
    \\in [0.5, 1]`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_elastic_in_out(0.2)
    0.007331378299120235
    >>> d3.ease_elastic_in_out.amplitude(2)(0.2)
    0.007331378299120243
    >>> d3.ease_elastic_in_out.period(0.5)(0.2)
    0.002265520486619852
    >>> d3.ease_elastic_in_out.amplitude(2).period(0.5)(0.2)
    0.014342340186712696
    """

    def __init__(self, amplitude: float, period: float):
        self._amplitude = max(1, amplitude)
        self._period = period / TAU
        self._s = asin(1 / self._amplitude) * self._period

    def __call__(self, t: float) -> float:
        """
        Symmetric elastic easing; scales :func:`d3.ease_elastic_in
        <detroit.ease_elastic_in>` :math:`\\forall t \\in [0, 0.5]` and
        :func:`d3.ease_elastic_out <detroit.ease_elastic_out>` :math:`\\forall
        t \\in [0.5, 1]`.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_elastic_in_out(0.2)
        0.007331378299120235
        >>> d3.ease_elastic_in_out.amplitude(2)(0.2)
        0.007331378299120243
        >>> d3.ease_elastic_in_out.period(0.5)(0.2)
        0.002265520486619852
        >>> d3.ease_elastic_in_out.amplitude(2).period(0.5)(0.2)
        0.014342340186712696
        """
        t = t * 2 - 1
        if t < 0:
            return self._amplitude * tqmt(-t) * sin((self._s - t) / self._period) * 0.5
        else:
            return (
                2 - self._amplitude * tqmt(t) * sin((self._s + t) / self._period)
            ) * 0.5

    def amplitude(self, amplitude: float) -> ElasticInOut:
        """
        Returns a new easing function with :code:`amplitude` value as
        parameter.

        Parameters
        ----------
        amplitude : float
            Amplitude value

        Returns
        -------
        ElasticInOut
            New easing function
        """
        return ElasticInOut(amplitude, self._period * TAU)

    def period(self, period: float) -> ElasticInOut:
        """
        Returns a new easing function with :code:`period` value as parameter.

        Parameters
        ----------
        period : float
            Amplitude value

        Returns
        -------
        ElasticInOut
            New easing function
        """
        return ElasticInOut(self._amplitude, period)


ease_elastic_in_out = ElasticInOut(AMPLITUDE, PERIOD)



from __future__ import annotations

import math
from bisect import bisect
from statistics import quantiles
from typing import Any, Generic, overload

from ..types import Number, T
from .init import init_range


class ScaleQuantile(Generic[T]):
    """
    Quantile scales map a sampled input domain to a discrete range.
    The domain is considered continuous and thus the scale will accept
    any reasonable input value; however, the domain is specified as a
    discrete set of sample values. The number of values in (the cardinality
    of) the output range determines the number of quantiles that will be
    computed from the domain. To compute the quantiles, the domain is sorted,
    and treated as a population of discrete values; see quantile.
    """

    def __init__(self):
        self._domain = []
        self._range_vals = []
        self._thresholds = []
        self._unknown = None

    def rescale(self):
        n = max(1, len(self._range_vals))
        self._thresholds = quantiles(self._domain, n=n, method="inclusive")
        return self

    def __call__(self, x: Number) -> T:
        """
        Given a value in the input domain, returns the
        corresponding value in the output range.

        Parameters
        ----------
        x : Number
            Input value

        Returns
        -------
        T
            Output value
        """
        if x is None or isinstance(x, float) and math.isnan(x):
            return self._unknown
        return self._range_vals[bisect(self._thresholds, x)]

    def invert_extent(self, y: T) -> Number:
        """
        Returns the extent of values in the domain :math:`[x_0, x_1]`
        for the corresponding value in the range: the inverse of quantile.

        Parameters
        ----------
        y : T
            Input value

        Returns
        -------
        Number
            Output value
        """
        if y not in self._range_vals:
            return [math.nan, math.nan]
        i = self._range_vals.index(y)
        return (
            [None, None]
            if i < 0
            else [
                self._thresholds[i - 1] if i > 0 else self._domain[0],
                self._thresholds[i] if i < len(self._thresholds) else self._domain[-1],
            ]
        )

    def set_domain(self, domain: list[Number]) -> ScaleQuantile:
        """
        Sets the domain of the quantile scale to the specified
        set of discrete numeric values .

        Parameters
        ----------
        domain : list[Number]
            Domain

        Returns
        -------
        ScaleQuantile
            Itself
        """
        self._domain.clear()
        for d in domain:
            if isinstance(d, str):
                d = float(d)
            if d is not None and not (isinstance(d, float) and math.isnan(d)):
                self._domain.append(d)
        self._domain = sorted(self._domain)
        return self.rescale()

    def get_domain(self) -> list[Number]:
        return self._domain.copy()

    def set_range(self, range_vals: list[T]) -> ScaleQuantile:
        """
        Sets the discrete values in the range.

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        ScaleQuantile
            Itself
        """
        self._range_vals = list(range_vals)
        return self.rescale()

    def get_range(self) -> list[T]:
        return self._range_vals.copy()

    def set_unknown(self, unknown: Any) -> ScaleQuantile:
        """
        Sets the scale's unknown value.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        ScaleQuantile
            Itself
        """
        self._unknown = unknown
        return self

    def get_unknown(self) -> Any:
        return self._unknown

    def get_quantiles(self):
        return self._thresholds.copy()

    def copy(self):
        return (
            ScaleQuantile()
            .set_domain(self._domain)
            .set_range(self._range_vals)
            .set_unknown(self._unknown)
        )

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


@overload
def scale_quantile() -> ScaleQuantile[T]: ...


@overload
def scale_quantile(range_vals: list[T]) -> ScaleQuantile[T]: ...


@overload
def scale_quantile(domain: list[Number], range_vals: list[T]) -> ScaleQuantile[T]: ...


def scale_quantile(*args):
    """
    Builds a new quantile scale with the specified domain and range.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    range_vals : list[T]
        Array of values

    Returns
    -------
    ScaleQuantile[T]
        Scale object

    Examples
    --------

    >>> scale = d3.scale_quantile().set_domain([3, 6, 7, 8, 8, 10, 13, 15, 16, 20])
    >>> scale = scale.set_range([0, 1, 2, 3])
    >>> for x in range(21):
    ...     print(x, scale(x))
    ...
    ...
    0 0
    1 0
    2 0
    3 0
    4 0
    5 0
    6 0
    7 0
    8 1
    9 2
    10 2
    11 2
    12 2
    13 2
    14 2
    15 3
    16 3
    17 3
    18 3
    19 3
    20 3
    """
    scale = ScaleQuantile()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



def format_decimal(x, _):
    x = round(x)
    if abs(x) >= 1e21:
        return str(x).replace(",", "")
    return str(x)


def format_decimal_parts(x, p):
    try:
        if p is not None:
            x_str = f"{x:.{p - 1}e}"
        else:
            x_str = f"{x:e}"

        i = x_str.find("e")
        if i < 0:
            return None  # NaN, ±Infinity

        coefficient = x_str[:i]
        if len(coefficient) > 1:
            coefficient = coefficient[0] + coefficient[2:]  # Remove decimal point

        return [coefficient, int(x_str[i + 1 :])]
    except (ValueError, TypeError):
        return None



import locale
from collections.abc import Callable
from datetime import datetime

from ..types import Formatter


def time_format(specifier: str) -> Formatter[str]:
    """
    Returns a formatter function to change a datetime into string

    Parameters
    ----------
    specifier : str
        Specifier string

    Returns
    -------
    Formatter[str]
        Formatter function which returns a string

    Examples
    --------

    >>> from datetime import datetime
    >>> d = datetime(2004, 6, 8, 12, 10)
    >>> d3.time_format("%Y")(d)
    '2004'
    """

    def formatter(date: datetime):
        return date.strftime(specifier)

    return formatter


def time_parse(specifier: str) -> Formatter[datetime]:
    """
    Returns a formatter function to change a string into datetime

    Parameters
    ----------
    specifier : str
        Specifier string

    Returns
    -------
    Formatter[datetime]
        Formatter function which returns a datetime

    Examples
    --------

    >>> d3.time_parse("%Y")("2004")
    datetime.datetime(2004, 1, 1, 0, 0)
    """

    def formatter(string: str):
        return datetime.strptime(string, specifier)

    return formatter


def iso_format(date: datetime) -> str:
    """
    Formats the date into

    Parameters
    ----------
    date : datetime
        Input date

    Returns
    -------
    str
        Iso string

    Examples
    --------

    >>> from datetime import datetime
    >>> d = datetime(2004, 6, 8, 12, 10)
    >>> d3.iso_format(d)
    '2004-06-08T12:10:00'
    """
    return date.isoformat()


def iso_parse(string: str) -> datetime:
    """
    Parses the iso string into datetime.

    Parameters
    ----------
    string : str
        Iso string

    Returns
    -------
    datetime
        Output date

    Examples
    --------

    >>> d3.iso_parse('2004-06-08T12:10:00')
    datetime.datetime(2004, 6, 8, 12, 10)
    """
    return datetime.fromisoformat(string)


def time_format_locale(language: str) -> Callable[[str], Formatter[str]]:
    """
    Change the locale language and return a formatter function
    (see :func:`d3.time_format <time_format>`).

    Parameters
    ----------
    language : str
        Language

    Returns
    -------
    Callable[[str], Formatter[str]]
        Function which takes a specifier string and returns a formatter function
        which itself returns a string

    Examples
    --------

    >>> f = time_format_locale("en_US.UTF-8")("%B %d, %Y")
    >>> f(datetime.now())
    'October 18, 2024'
    """
    locale.setlocale(locale.LC_TIME, language)

    def formatter(specifier: str):
        return time_format(specifier)

    return formatter



import math
from statistics import quantiles


def threshold_freedman_diaconis(
    values: list[float | None], mini: float, maxi: float
) -> int:
    """
    Returns the number of bins according to the Freedman–Diaconis rule; the
    input values must be numbers.

    Parameters
    ----------
    values : list[float | None]
        Input values
    mini : float
        Minimum value
    maxi : float
        Maximum value

    Returns
    -------
    int
        Number of bins according to the Freedman-Diaconis rule
    """
    values = [v for v in values if v is not None]
    c = len(values)
    if len(values) <= 1:
        d = 0
    else:
        q = quantiles(values, n=4, method="inclusive")
        d = q[2] - q[0]
    return math.ceil((maxi - mini) / (2 * d * (c ** (-1 / 3)))) if c and d else 1



from math import isnan

from ...types import T
from ..series import Series
from .none import offset_none


def offset_wiggle(series: list[Series[T]], order: list[int]):
    """
    Shifts the baseline so as to minimize the weighted wiggle of layers. This
    offset is recommended for streamgraphs in conjunction with
    :func:`d3.stack_order_inside_out <stack_order_inside_out>`.

    Parameters
    ----------
    series : list[Series[T]]
        List of series
    order : list[int]
        List of ordered indices
    """
    n = len(series)
    if n == 0:
        return
    s0 = series[order[0]]
    m = len(s0)
    if m == 0:
        return
    y = 0
    for j in range(1, m):
        s1 = 0
        s2 = 0
        for i in range(n):
            si = series[order[i]]
            x1 = si[j][1]
            x2 = si[j - 1][1]
            sij0 = 0.0 if isnan(x1) else x1
            sij1 = 0.0 if isnan(x2) else x2
            s3 = (sij0 - sij1) / 2
            for k in range(i):
                sk = series[order[k]]
                x1 = sk[j][1]
                x2 = sk[j - 1][1]
                skj0 = 0.0 if isnan(x1) else x1
                skj1 = 0.0 if isnan(x2) else x2
                s3 += skj0 - skj1
            s1 += sij0
            s2 += s3 * sij0
        s0[j - 1][0] = y
        s0[j - 1][1] += s0[j - 1][0]
        if s1 and not isnan(s1):
            y -= s2 / s1
    s0[j][0] = y
    s0[j][1] += s0[j][0]
    offset_none(series, order)



from .exponent import exponent


def precision_prefix(step: int | float, value: int | float) -> int:
    """
    Returns a suggested decimal precision for use with :code:`d3.format_prefix`
    given the specified numeric step and reference value. The step represents
    the minimum absolute difference between values that will be formatted,
    and value determines which SI prefix will be used.

    Parameters
    ----------
    step : int | float
        Specific numeric value
    value : int | float
        Reference value

    Returns
    -------
    int
        Decimal precision

    Examples
    --------

    >>> p = d3.precision_prefix(1e5, 1.3e6)
    >>> f = d3.format_prefix(f".{p}", 1.3e6)
    >>> f(1.1e6)
    '1.1M'
    >>> f(1.2e6)
    '1.2M'
    >>> f(1.3e6)
    '1.3M'
    """
    return max(0, max(-8, min(8, exponent(value) // 3)) * 3 - exponent(abs(step)))



from collections.abc import Callable

from ..types import T, U, V
from .argpass import argpass


def nest(
    values: list[U],
    map_function: Callable[[dict], list],
    reduce_function: Callable[[list[U]], dict],
    keys: list[Callable[[U, int, list[U]], V]],
) -> list:
    if len(keys) == 0:
        raise ValueError("At least one key must be declared.")

    def regroup(values, i):
        if i >= len(keys):
            return reduce_function(values)
        groups = {}
        keyof = argpass(keys[i])
        i += 1
        index = -1
        for value in values:
            index += 1
            key = keyof(value, index, values)
            if group := groups.get(key):
                group.append(value)
            else:
                groups[key] = [value]
        for key, values in groups.items():
            groups[key] = regroup(values, i)
        return map_function(groups)

    return regroup(values, 0)


def identity(x: T) -> T:
    return x


def unique(values: list[T]) -> T:
    if len(values) != 1:
        raise IndexError("Duplicate key")
    return values[0]


def array_from(groups: dict) -> list:
    return list(groups.items())


def index(
    values: list[U],
    *keys: Callable[[U, int, list[U]], V],
) -> dict[V, U]:
    """
    Groups and reduces the specified list of values into a nested dictionary.
    The reducer extracts the first element from each group.

    Parameters
    ----------
    values : list[U]
        List of values
    keys : Callable[[U, int, list[U]], V]
        List of functions which take in arguments data, index and list of data
        and returns the key value of the data.

    Returns
    -------
    dict[V, U]
        Nested dictionary.

    Examples
    --------

    >>> data = [
    ...     {"id": 0, "value": 10},
    ...     {"id": 0, "value": 20},
    ...     {"id": 1, "value": 3},
    ... ]
    >>> d3.index(data, lambda d: d["value"])
    {10: {'id': 0, 'value': 10}, 20: {'id': 0, 'value': 20}, 3: {'id': 1, 'value': 3}}
    """
    return nest(values, identity, unique, keys)


def indexes(
    values: list[U],
    *keys: Callable[[U, int, list[U]], V],
) -> list[tuple[V, U]]:
    """
    Equivalent to :func:`d3.index <index>`, returns a list of collections [key, array of
    values]. The reducer extracts the first element from each group.

    Parameters
    ----------
    values : list[U]
        List of values
    *keys : Callable[[U, int, list[U]], V]
        List of functions which take in arguments data, index and list of data
        and returns the key value of the data.

    Returns
    -------
    list[tuple[V, U]]
        Nested list of collections [key; array of values].

    Examples
    --------

    >>> data = [
    ...     {"id": 0, "value": 10},
    ...     {"id": 0, "value": 20},
    ...     {"id": 1, "value": 3},
    ... ]
    >>> d3.indexes(data, lambda d: d["value"])
    [(10, {'id': 0, 'value': 10}), (20, {'id': 0, 'value': 20}), (3, {'id': 1, 'value': 3})]
    """
    return nest(values, array_from, unique, keys)


def group(
    values: list[U],
    *keys: Callable[[U, int, list[U]], V],
) -> dict[V, list[U]]:
    """
    Groups the specified list of values into a nested dictionary given keys.

    Parameters
    ----------
    values : list[U]
        List of values
    *keys : Callable[[U, int, list[U]], V]
        List of functions which take in arguments data, index and list of data
        and returns the key value of the data.

    Returns
    -------
    dict[V, list[U]]
        Nested dictionary

    Examples
    --------

    >>> data = [
    ...     {"id": 0, "value": 10},
    ...     {"id": 0, "value": 20},
    ...     {"id": 1, "value": 3},
    ... ]
    >>> d3.group(data, lambda d: d["id"])
    {0: [{'id': 0, 'value': 10}, {'id': 0, 'value': 20}], 1: [{'id': 1, 'value': 3}]}
    """
    return nest(values, identity, identity, keys)


def groups(
    values: list[U],
    *keys: Callable[[U, int, list[U]], V],
) -> list[tuple[V, list[U]]]:
    """
    Equivalent to :func:`d3.group <group>`, returns a list of collections [key; array of
    values].

    Parameters
    ----------
    values : list[U]
        List of values
    *keys : Callable[[U, int, list[U]], V]
        List of functions which take in arguments data, index and list of data
        and returns the key value of the data.

    Returns
    -------
    list[tuple[V, list[U]]]
        Nested list of collections [key; array of values].

    Examples
    --------

    >>> data = [
    ...     {"id": 0, "value": 10},
    ...     {"id": 0, "value": 20},
    ...     {"id": 1, "value": 3},
    ... ]
    >>> d3.groups(data, lambda d: d["id"])
    [(0, [{'id': 0, 'value': 10}, {'id': 0, 'value': 20}]), (1, [{'id': 1, 'value': 3}])]
    """
    return nest(values, array_from, identity, keys)


def rollup(
    values: list[T],
    reduce_function: Callable[[list[T]], V],
    *keys: Callable[[T, int, list[T]], U],
) -> dict[U, V]:
    """
    Groups and reduces the specified list of values into a nested dictionary.

    Parameters
    ----------
    values : list[T]
        List of values
    reduce_function : Callable[[list[T]], V]
        Reducer function
    *keys : Callable[[T, int, list[T]], U]
        List of functions which take in arguments data, index and list of data
        and returns the key value of the data.

    Returns
    -------
    dict[U, V]
        Nested dictionary

    Examples
    --------

    >>> data = [
    ...     {"id": 0, "value": 10},
    ...     {"id": 0, "value": 20},
    ...     {"id": 1, "value": 3},
    ... ]
    >>> d3.rollup(
    ...     data,
    ...     lambda values: sum([d["value"] for d in values]),
    ...     lambda d: d["id"],
    ... )
    {0: 30, 1: 3}
    """
    return nest(values, identity, reduce_function, keys)


def rollups(
    values: list[T],
    reduce_function: Callable[[list[T]], V],
    *keys: Callable[[T, int, list[T]], U],
) -> list[tuple[U, V]]:
    """
    Equivalent to `d3.rollup`, returns a list of collections [key; array of
    values].

    Parameters
    ----------
    values : list[T]
        List of values
    reduce_function : Callable[[list[T]], V]
        Reducer function
    *keys : Callable[[T, int, list[T]], U]
        List of functions which take in arguments data, index and list of data
        and returns the key value of the data.

    Returns
    -------
    list[tuple[U, V]]
        Nested list of collections [key; array of values].

    Examples
    --------

    >>> data = [
    ...     {"id": 0, "value": 10},
    ...     {"id": 0, "value": 20},
    ...     {"id": 1, "value": 3},
    ... ]
    >>> d3.rollups(
    ...     data,
    ...     lambda values: sum([d["value"] for d in values]),
    ...     lambda d: d["id"],
    ... )
    [(0, 30), (1, 3)]
    """
    return nest(values, array_from, reduce_function, keys)



from __future__ import annotations

from collections.abc import Callable
from math import isnan, sqrt

from ..array import argpass
from ..types import SimulationNode, SimulationNodeFunction
from .constant import constant


class ForceRadial:
    def __init__(self, radius: SimulationNodeFunction[float], x: float, y: float):
        self._radius = radius
        self._x = x
        self._y = y
        self._strength = constant(0.1)
        self._strengths = None
        self._radiuses = None

    def __call__(self, alpha: float):
        for i, node in enumerate(self._nodes):
            dx = (node["x"] - self._x) or 1e-6
            dy = (node["y"] - self._y) or 1e-6
            r = sqrt(dx * dx + dy * dy)
            k = (self._radiuses[i] - r) * self._strengths[i] * alpha / r
            node["vx"] += dx * k
            node["vy"] += dy * k

    def _initialize(self):
        if self._nodes is None:
            return

        n = len(self._nodes)
        self._strengths = [None] * n
        self._radiuses = [None] * n

        for i, node in enumerate(self._nodes):
            self._radiuses[i] = r = self._radius(node, i, self._nodes)
            self._strengths[i] = 0 if isnan(r) else self._strength(node, i, self._nodes)

    def initialize(self, nodes: list[SimulationNode], random: Callable[[None], float]):
        self._nodes = nodes
        self._initialize()

    def set_strength(
        self, strength: SimulationNodeFunction[float] | float
    ) -> ForceRadial:
        """
        Sets the strength accessor to the specified number or function,
        re-evaluates the strength accessor for each node, and returns this
        force. The strength determines how much to increment the node's x- and
        y-velocity.

        For example, a value of :math:`0.1` indicates that the node should move
        a tenth of the way from its current position to the closest point on
        the circle with each application. Higher values moves nodes more
        quickly to the target position, often at the expense of other forces or
        constraints. A value outside the range :math:`[0,1]` is not
        recommended.

        Parameters
        ----------
        strength : SimulationNodeFunction[float] | float
            Strength function or constant value. If it is a function, it takes
            the following arguments:

            * **node** (:code:`SimulationNode`) - the node element
            * **i** (:code:`int`) - the index of the node
            * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

            It returns the strength value (:code:`float`)

        Returns
        -------
        ForceRadial
            Itself
        """
        if callable(strength):
            self._strength = argpass(strength)
        else:
            self._strength = constant(strength)
        self._initialize()
        return self

    def set_radius(self, radius: SimulationNodeFunction[float] | float) -> ForceRadial:
        """
        Sets the circle radius to the specified number or function,
        re-evaluates the radius accessor for each node, and returns this force.

        Parameters
        ----------
        radius : SimulationNodeFunction[float] | float
            Radius function or constant value. If it is a function, it takes
            the following arguments:

            * **node** (:code:`SimulationNode`) - the node element
            * **i** (:code:`int`) - the index of the node
            * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

            It returns a radius value (:code:`float`).

        Returns
        -------
        ForceRadial
            Force object
        """
        if callable(radius):
            self._radius = argpass(radius)
        else:
            self._radius = constant(radius)
        self._initialize()
        return self

    def x(self, x: float) -> ForceRadial:
        """
        Sets the x-coordinate of the circle center to the specified number and
        returns this force.

        Parameters
        ----------
        x : float
            X-coordinate of circle center

        Returns
        -------
        ForceRadial
            Itself
        """
        self._x = x
        return self

    def y(self, y: float) -> ForceRadial:
        """
        Sets the y-coordinate of the circle center to the specified number and
        returns this force.

        Parameters
        ----------
        y : float
            y-coordinate of circle center

        Returns
        -------
        ForceRadial
            Itself
        """
        self._y = y
        return self

    def get_strength(self) -> SimulationNodeFunction[float]:
        return self._strength

    def get_radius(self) -> SimulationNodeFunction[float]:
        return self._radius

    def get_x(self) -> float:
        return self._x

    def get_y(self) -> float:
        return self._y


def force_radial(
    radius: SimulationNodeFunction[float] | float,
    x: float | None = None,
    y: float | None = None,
) -> ForceRadial:
    """
    Creates a new position force towards a circle of the specified radius
    centered at :math`(x,y)`. If x and y are not specified, they default to
    :code:`(0,0)`.

    Parameters
    ----------
    radius : SimulationNodeFunction[float] | float
        Radius function or constant value. If it is a function, it takes
        the following arguments:

        * **node** (:code:`SimulationNode`) - the node element
        * **i** (:code:`int`) - the index of the node
        * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

        It returns a radius value (:code:`float`).

    x : float | None
        X-coordinate of circle center
    y : float | None
        Y-coordinate of circle center

    Returns
    -------
    ForceRadial
        Force object
    """
    if not callable(radius):
        radius = constant(radius)
    else:
        radius = argpass(radius)
    if x is None:
        x = 0
    if y is None:
        y = 0

    return ForceRadial(radius, x, y)



from .basis import curve_basis
from .basis_closed import curve_basis_closed
from .basis_open import curve_basis_open
from .bump import curve_bump_radial, curve_bump_x, curve_bump_y
from .bundle import curve_bundle
from .cardinal import curve_cardinal
from .cardinal_closed import curve_cardinal_closed
from .cardinal_open import curve_cardinal_open
from .catmull_rom import curve_catmull_rom
from .catmull_rom_closed import curve_catmull_rom_closed
from .catmull_rom_open import curve_catmull_rom_open
from .common import Curve
from .linear import curve_linear
from .linear_closed import curve_linear_closed
from .monotone import curve_monotone_x, curve_monotone_y
from .natural import curve_natural
from .radial import curve_radial, curve_radial_linear
from .step import curve_step, curve_step_after, curve_step_before

__all__ = [
    "Curve",
    "curve_basis",
    "curve_basis_closed",
    "curve_basis_open",
    "curve_bump_radial",
    "curve_bump_x",
    "curve_bump_y",
    "curve_bundle",
    "curve_cardinal",
    "curve_cardinal_closed",
    "curve_cardinal_open",
    "curve_catmull_rom",
    "curve_catmull_rom_closed",
    "curve_catmull_rom_open",
    "curve_linear",
    "curve_linear_closed",
    "curve_monotone_x",
    "curve_monotone_y",
    "curve_natural",
    "curve_radial",
    "curve_radial_linear",
    "curve_step",
    "curve_step_after",
    "curve_step_before",
]



from __future__ import annotations

import math
import re
from typing import overload

DARKER = 0.7
BRIGHTER = 1 / DARKER

RE_I = re.compile(r"\s*([+-]?\d+)\s*")
RE_N = re.compile(r"\s*([+-]?(?:\d*\.)?\d+(?:[eE][+-]?\d+)?)\s*")
RE_P = re.compile(r"\s*([+-]?(?:\d*\.)?\d+(?:[eE][+-]?\d+)?)%\s*")
RE_HEX = re.compile(r"^#([0-9a-f]{3,8})$")
RE_RGB_INTEGER = re.compile(rf"^rgb\({RE_I.pattern},{RE_I.pattern},{RE_I.pattern}\)$")
RE_RGB_PERCENT = re.compile(rf"^rgb\({RE_P.pattern},{RE_P.pattern},{RE_P.pattern}\)$")
RE_RGBA_INTEGER = re.compile(
    rf"^rgba\({RE_I.pattern},{RE_I.pattern},{RE_I.pattern},{RE_N.pattern}\)$"
)
RE_RGBA_PERCENT = re.compile(
    rf"^rgba\({RE_P.pattern},{RE_P.pattern},{RE_P.pattern},{RE_N.pattern}\)$"
)
RE_HSL_PERCENT = re.compile(rf"^hsl\({RE_N.pattern},{RE_P.pattern},{RE_P.pattern}\)$")
RE_HSLA_PERCENT = re.compile(
    rf"^hsla\({RE_N.pattern},{RE_P.pattern},{RE_P.pattern},{RE_N.pattern}\)$"
)


NAMED = {
    "aliceblue": 0xF0F8FF,
    "antiquewhite": 0xFAEBD7,
    "aqua": 0x00FFFF,
    "aquamarine": 0x7FFFD4,
    "azure": 0xF0FFFF,
    "beige": 0xF5F5DC,
    "bisque": 0xFFE4C4,
    "black": 0x000000,
    "blanchedalmond": 0xFFEBCD,
    "blue": 0x0000FF,
    "blueviolet": 0x8A2BE2,
    "brown": 0xA52A2A,
    "burlywood": 0xDEB887,
    "cadetblue": 0x5F9EA0,
    "chartreuse": 0x7FFF00,
    "chocolate": 0xD2691E,
    "coral": 0xFF7F50,
    "cornflowerblue": 0x6495ED,
    "cornsilk": 0xFFF8DC,
    "crimson": 0xDC143C,
    "cyan": 0x00FFFF,
    "darkblue": 0x00008B,
    "darkcyan": 0x008B8B,
    "darkgoldenrod": 0xB8860B,
    "darkgray": 0xA9A9A9,
    "darkgreen": 0x006400,
    "darkgrey": 0xA9A9A9,
    "darkkhaki": 0xBDB76B,
    "darkmagenta": 0x8B008B,
    "darkolivegreen": 0x556B2F,
    "darkorange": 0xFF8C00,
    "darkorchid": 0x9932CC,
    "darkred": 0x8B0000,
    "darksalmon": 0xE9967A,
    "darkseagreen": 0x8FBC8F,
    "darkslateblue": 0x483D8B,
    "darkslategray": 0x2F4F4F,
    "darkslategrey": 0x2F4F4F,
    "darkturquoise": 0x00CED1,
    "darkviolet": 0x9400D3,
    "deeppink": 0xFF1493,
    "deepskyblue": 0x00BFFF,
    "dimgray": 0x696969,
    "dimgrey": 0x696969,
    "dodgerblue": 0x1E90FF,
    "firebrick": 0xB22222,
    "floralwhite": 0xFFFAF0,
    "forestgreen": 0x228B22,
    "fuchsia": 0xFF00FF,
    "gainsboro": 0xDCDCDC,
    "ghostwhite": 0xF8F8FF,
    "gold": 0xFFD700,
    "goldenrod": 0xDAA520,
    "gray": 0x808080,
    "green": 0x008000,
    "greenyellow": 0xADFF2F,
    "grey": 0x808080,
    "honeydew": 0xF0FFF0,
    "hotpink": 0xFF69B4,
    "indianred": 0xCD5C5C,
    "indigo": 0x4B0082,
    "ivory": 0xFFFFF0,
    "khaki": 0xF0E68C,
    "lavender": 0xE6E6FA,
    "lavenderblush": 0xFFF0F5,
    "lawngreen": 0x7CFC00,
    "lemonchiffon": 0xFFFACD,
    "lightblue": 0xADD8E6,
    "lightcoral": 0xF08080,
    "lightcyan": 0xE0FFFF,
    "lightgoldenrodyellow": 0xFAFAD2,
    "lightgray": 0xD3D3D3,
    "lightgreen": 0x90EE90,
    "lightgrey": 0xD3D3D3,
    "lightpink": 0xFFB6C1,
    "lightsalmon": 0xFFA07A,
    "lightseagreen": 0x20B2AA,
    "lightskyblue": 0x87CEFA,
    "lightslategray": 0x778899,
    "lightslategrey": 0x778899,
    "lightsteelblue": 0xB0C4DE,
    "lightyellow": 0xFFFFE0,
    "lime": 0x00FF00,
    "limegreen": 0x32CD32,
    "linen": 0xFAF0E6,
    "magenta": 0xFF00FF,
    "maroon": 0x800000,
    "mediumaquamarine": 0x66CDAA,
    "mediumblue": 0x0000CD,
    "mediumorchid": 0xBA55D3,
    "mediumpurple": 0x9370DB,
    "mediumseagreen": 0x3CB371,
    "mediumslateblue": 0x7B68EE,
    "mediumspringgreen": 0x00FA9A,
    "mediumturquoise": 0x48D1CC,
    "mediumvioletred": 0xC71585,
    "midnightblue": 0x191970,
    "mintcream": 0xF5FFFA,
    "mistyrose": 0xFFE4E1,
    "moccasin": 0xFFE4B5,
    "navajowhite": 0xFFDEAD,
    "navy": 0x000080,
    "oldlace": 0xFDF5E6,
    "olive": 0x808000,
    "olivedrab": 0x6B8E23,
    "orange": 0xFFA500,
    "orangered": 0xFF4500,
    "orchid": 0xDA70D6,
    "palegoldenrod": 0xEEE8AA,
    "palegreen": 0x98FB98,
    "paleturquoise": 0xAFEEEE,
    "palevioletred": 0xDB7093,
    "papayawhip": 0xFFEFD5,
    "peachpuff": 0xFFDAB9,
    "peru": 0xCD853F,
    "pink": 0xFFC0CB,
    "plum": 0xDDA0DD,
    "powderblue": 0xB0E0E6,
    "purple": 0x800080,
    "rebeccapurple": 0x663399,
    "red": 0xFF0000,
    "rosybrown": 0xBC8F8F,
    "royalblue": 0x4169E1,
    "saddlebrown": 0x8B4513,
    "salmon": 0xFA8072,
    "sandybrown": 0xF4A460,
    "seagreen": 0x2E8B57,
    "seashell": 0xFFF5EE,
    "sienna": 0xA0522D,
    "silver": 0xC0C0C0,
    "skyblue": 0x87CEEB,
    "slateblue": 0x6A5ACD,
    "slategray": 0x708090,
    "slategrey": 0x708090,
    "snow": 0xFFFAFA,
    "springgreen": 0x00FF7F,
    "steelblue": 0x4682B4,
    "tan": 0xD2B48C,
    "teal": 0x008080,
    "thistle": 0xD8BFD8,
    "tomato": 0xFF6347,
    "turquoise": 0x40E0D0,
    "violet": 0xEE82EE,
    "wheat": 0xF5DEB3,
    "white": 0xFFFFFF,
    "whitesmoke": 0xF5F5F5,
    "yellow": 0xFFFF00,
    "yellowgreen": 0x9ACD32,
}


class Color:
    """
    Color base class used by :code:`RGB`, :code:`HSL`, :code:`Cubehelix`,
    :code:`LAB` and :code:`HCL`.
    """

    def format_hex(self) -> str:
        """
        Returns the color formatted as hex color.

        Returns
        -------
        str
            Hex color

        Examples
        --------

        >>> d3.rgb(128, 250, 102, 0.2).format_hex()
        '#80fa66'
        """
        return self.rgb().format_hex()

    def format_hex_8(self) -> str:
        """
        Returns the color formatted as hex color with alpha channel (opacity).

        Returns
        -------
        str
            Hex color

        Examples
        --------

        >>> d3.rgb(128, 250, 102, 0.2).format_hex_8()
        '#80fa6633'
        """
        return self.rgb().format_hex_8()

    def format_hsl(self) -> str:
        """
        Returns the color formatted as HSL color.

        Returns
        -------
        str
            HSL color

        Examples
        --------

        >>> d3.rgb(128, 250, 102, 0.2).format_hsl()
        'hsla(109.45945945945947, 93.67088607594937%, 69.01960784313725%, 0.2)'
        """
        return hsl_convert(self).format_hsl()

    def format_rgb(self) -> str:
        """
        Returns the color formatted as RGB color.

        Returns
        -------
        str
            RGB color

        Examples
        --------

        >>> d3.rgb(128, 250, 102, 0.2).format_rgb()
        'rgba(128, 250, 102, 0.2)'
        """
        return self.rgb().format_rgb()

    def __str__(self):
        return self.format_rgb()


def color(format: str) -> RGB | HSL | None:
    """
    Parses the specified CSS Color specifier string,
    returning an RGB or HSL color.
    If the specifier was not valid, None is returned.

    Parameters
    ----------
    format : str
        Specifier

    Returns
    -------
    RGB | HSL | None
        Formatted color

    Examples
    --------

    >>> d3.color("#2e65ffff")
    RGB(r=46, g=101, b=255, opacity=1.0)
    >>> d3.color("rgb(108, 10, 204)")
    RGB(r=108, g=10, b=204, opacity=1.0)
    >>> d3.color("hsl(210.4, 90%, 70%)")
    HSL(h=210.4, s=0.9, l=0.7, opacity=1.0)
    >>> d3.color("bad") is None
    True
    """
    format = format.strip().lower()
    if format == "transparent":
        return RGB(math.nan, math.nan, math.nan, 0)
    elif m := RE_HEX.match(format):
        length = len(m.group(1))
        m = int(m.group(1), 16)
        if length == 6:
            return rgbn(m)  # #ff0000
        elif length == 3:
            return RGB(
                (m >> 8 & 0xF) | (m >> 4 & 0xF0),
                (m >> 4 & 0xF) | (m & 0xF0),
                ((m & 0xF) << 4) | (m & 0xF),
                1,
            )  # #f00
        elif length == 8:
            return rgba(
                m >> 24 & 0xFF, m >> 16 & 0xFF, m >> 8 & 0xFF, (m & 0xFF) / 0xFF
            )  # #ff000000
        elif length == 4:
            return rgba(
                (m >> 12 & 0xF) | (m >> 8 & 0xF0),
                (m >> 8 & 0xF) | (m >> 4 & 0xF0),
                (m >> 4 & 0xF) | (m & 0xF0),
                (((m & 0xF) << 4) | (m & 0xF)) / 0xFF,
            )  # #f000
        return None  # invalid hex
    elif m := RE_RGB_INTEGER.match(format):
        return RGB(
            int(m.group(1)), int(m.group(2)), int(m.group(3)), 1
        )  # rgb(255, 0, 0)
    elif m := RE_RGB_PERCENT.match(format):
        return RGB(
            float(m.group(1)) * 255 / 100,
            float(m.group(2)) * 255 / 100,
            float(m.group(3)) * 255 / 100,
            1,
        )  # rgb(100%, 0%, 0%)
    elif m := RE_RGBA_INTEGER.match(format):
        return rgba(
            int(m.group(1)), int(m.group(2)), int(m.group(3)), float(m.group(4))
        )  # rgba(255, 0, 0, 1)
    elif m := RE_RGBA_PERCENT.match(format):
        return rgba(
            float(m.group(1)) * 255 / 100,
            float(m.group(2)) * 255 / 100,
            float(m.group(3)) * 255 / 100,
            float(m.group(4)),
        )  # rgb(100%, 0%, 0%, 1)
    elif m := RE_HSL_PERCENT.match(format):
        return hsla(
            float(m.group(1)), float(m.group(2)) / 100, float(m.group(3)) / 100, 1
        )  # hsl(120, 50%, 50%)
    elif m := RE_HSLA_PERCENT.match(format):
        return hsla(
            float(m.group(1)),
            float(m.group(2)) / 100,
            float(m.group(3)) / 100,
            float(m.group(4)),
        )  # hsla(120, 50%, 50%, 1)
    elif format in NAMED:
        return rgbn(NAMED[format])
    return None


def rgbn(n):
    return RGB(n >> 16 & 0xFF, n >> 8 & 0xFF, n & 0xFF, 1)


def rgba(r, g, b, a):
    if a <= 0:
        r = g = b = math.nan
    return RGB(r, g, b, a)


def rgb_convert(obj):
    if not isinstance(obj, (Color, RGB, HSL)):
        obj = color(obj)
    if not obj:
        return RGB(0, 0, 0)
    obj = obj.rgb()
    return RGB(obj.r, obj.g, obj.b, obj.opacity)


def clampa(opacity):
    return 1 if math.isnan(opacity) else max(0, min(1, opacity))


def clampi(value):
    return 0 if math.isnan(value) else max(0, min(255, round(value) or 0))


def hex(value):
    value = clampi(value)
    return f"{'0' if value < 16 else ''}{value:x}"


def hsla(h, s, l, a):
    if a <= 0:
        h = s = l = math.nan
    elif l <= 0 or l >= 1:
        h = s = math.nan
    elif s <= 0:
        h = math.nan
    return HSL(h, s, l, a)


def hsl_convert(obj):
    if isinstance(obj, HSL):
        return HSL(obj.h, obj.s, obj.l, obj.opacity)
    if not isinstance(obj, Color):
        obj = color(obj)
    if not obj:
        return HSL(0, 0, 0)
    if isinstance(obj, HSL):
        return obj
    obj = obj.rgb()
    r = obj.r / 255
    g = obj.g / 255
    b = obj.b / 255
    min_val = min(r, g, b)
    max_val = max(r, g, b)
    h = math.nan
    s = max_val - min_val
    l = (max_val + min_val) / 2
    if s:
        if r == max_val:
            h = (g - b) / s + (g < b) * 6
        elif g == max_val:
            h = (b - r) / s + 2
        else:
            h = (r - g) / s + 4
        s /= l < 0.5 and (max_val + min_val) or (2 - max_val - min_val)
        h *= 60
    else:
        s = 0 if 0 < l < 1 else h
    return HSL(h, s, l, obj.opacity)


def clamph(value):
    value = (value or 0) % 360
    return value + 360 if value < 0 else value


def clampt(value):
    return max(0, min(1, value or 0))


def hsl2rgb(h, m1, m2):
    if h < 60:
        return (m1 + (m2 - m1) * h / 60) * 255
    elif h < 180:
        return m2 * 255
    elif h < 240:
        return (m1 + (m2 - m1) * (240 - h) / 60) * 255
    else:
        return m1 * 255


class RGB(Color):
    """
    RGB color format

    Parameters
    ----------
    r : int
        Red channel value
    g : int
        Green channel value
    b : int
        Blue channel value
    opacity : float
        Opacity value
    """

    def __init__(self, r: int, g: int, b: int, opacity: float = 1):
        self.r = float(r)
        self.g = float(g)
        self.b = float(b)
        self.opacity = float(opacity)

    def brighter(self, k: float | None = None) -> RGB:
        """
        Returns a brighter copy of this color.
        For example, if k is 1, steelblue in RGB color space becomes rgb(100, 186, 255).
        The parameter k controls how much brighter the returned color should be (in arbitrary units);
        if k is not specified, it defaults to 1. The behavior of this method is dependent
        on the implementing color space.

        Parameters
        ----------
        k : float | None
            Brightness coefficient

        Returns
        -------
        RGB
            Brighter RGB
        """
        k = BRIGHTER if k is None else BRIGHTER**k
        return RGB(self.r * k, self.g * k, self.b * k, self.opacity)

    def darker(self, k: float | None = None) -> RGB:
        """
        Returns a darker copy of this color.
        For example, if k is 1, steelblue in RGB color space becomes rgb(49, 91, 126).
        The parameter k controls how much darker the returned color should be (in arbitrary units);
        if k is not specified, it defaults to 1. The behavior of this method is dependent
        on the implementing color space.

        Parameters
        ----------
        k : float | None
            Darkness coefficient

        Returns
        -------
        RGB
            Darker RGB
        """
        k = DARKER if k is None else DARKER**k
        return RGB(self.r * k, self.g * k, self.b * k, self.opacity)

    def rgb(self) -> RGB:
        """
        Returns the RGB equivalent of this color

        Returns
        -------
        RGB
            RGB color format
        """
        return self

    def clamp(self) -> RGB:
        """
        Returns a new RGB color where the r, g, and b channels are clamped
        to the range [0, 255] and rounded to the nearest integer value, and
        the opacity is clamped to the range [0, 1].

        Returns
        -------
        RGB
            Clamped color
        """
        return RGB(clampi(self.r), clampi(self.g), clampi(self.b), clampa(self.opacity))

    def displayable(self) -> bool:
        """
        Returns :code:`True` if and only if the color is displayable on standard hardware.
        For example, this returns false for an RGB color if any channel value is
        less than zero or greater than 255 when rounded, or if the opacity is not
        in the range [0, 1].

        Returns
        -------
        bool
            Is displayable
        """
        return (
            (-0.5 <= self.r < 255.5)
            and (-0.5 <= self.g < 255.5)
            and (-0.5 <= self.b < 255.5)
            and (0 <= self.opacity <= 1)
        )

    def format_hex(self) -> str:
        """
        Returns a hexadecimal string representing this color in RGB space, such as #4682b4.
        If this color is not displayable, a suitable displayable color is returned instead.
        For example, RGB channel values greater than 255 are clamped to 255.

        Returns
        -------
        str
            Hex color representation
        """
        return f"#{hex(self.r)}{hex(self.g)}{hex(self.b)}"

    def format_hex_8(self) -> str:
        """
        Returns a hexadecimal string representing this color in RGBA space, such as #4682b4cc.
        If this color is not displayable, a suitable displayable color is returned instead.
        For example, RGB channel values greater than 255 are clamped to 255.

        Returns
        -------
        str
            Hex 8 color representation
        """
        return f"#{hex(self.r)}{hex(self.g)}{hex(self.b)}{hex((1 if math.isnan(self.opacity) else self.opacity) * 255)}"

    def format_rgb(self) -> str:
        """
        Returns a string representing this color according to the CSS Object Model specification,
        such as rgb(247, 234, 186) or rgba(247, 234, 186, 0.2). If this color is not displayable,
        a suitable displayable color is returned instead by clamping RGB channel values to the
        interval [0, 255].

        Returns
        -------
        str
            RGB color representation
        """
        a = clampa(self.opacity)
        return f"{'rgb(' if a == 1 else 'rgba('}{clampi(self.r)}, {clampi(self.g)}, {clampi(self.b)}{')' if a == 1 else f', {a})'}"

    def __repr__(self) -> str:
        return f"RGB(r={int(self.r)}, g={int(self.g)}, b={int(self.b)}, opacity={self.opacity})"


class HSL(Color):
    """
    HSL color format

    Parameters
    ----------
    h : float
        Hue channel value
    s : float
        Saturation channel value
    l : float
        Lightness channel value
    opacity : float
        Opacity value
    """

    def __init__(self, h: float, s: float, l: float, opacity: float = 1.0):
        self.h = float(h)
        self.s = float(s)
        self.l = float(l)
        self.opacity = float(opacity)

    def brighter(self, k: float | None = None) -> HSL:
        """
        Returns a brighter copy of this color.
        For example, if k is 1, steelblue in RGB color space becomes rgb(100, 186, 255).
        The parameter k controls how much brighter the returned color should be (in arbitrary units);
        if k is not specified, it defaults to 1. The behavior of this method is dependent
        on the implementing color space.

        Parameters
        ----------
        k : float | None
            Brightness coefficient

        Returns
        -------
        HSL
            Brighter HSL
        """
        k = BRIGHTER if k is None else BRIGHTER**k
        return HSL(self.h, self.s, self.l * k, self.opacity)

    def darker(self, k: float | None = None) -> HSL:
        """
        Returns a darker copy of this color.
        For example, if k is 1, steelblue in RGB color space becomes rgb(49, 91, 126).
        The parameter k controls how much darker the returned color should be (in arbitrary units);
        if k is not specified, it defaults to 1. The behavior of this method is dependent
        on the implementing color space.

        Parameters
        ----------
        k : float | None
            Darkness coefficient

        Returns
        -------
        HSL
            Darker HSL
        """
        k = DARKER if k is None else DARKER**k
        return HSL(self.h, self.s, self.l * k, self.opacity)

    def rgb(self) -> RGB:
        """
        Returns the RGB equivalent of this color

        Returns
        -------
        RGB
            RGB color format
        """
        h = self.h % 360
        s = 0 if math.isnan(h) or math.isnan(self.s) else self.s
        l = self.l
        m2 = l + (l if l < 0.5 else 1 - l) * s
        m1 = 2 * l - m2
        return RGB(
            hsl2rgb(h - 240 if h >= 240 else h + 120, m1, m2),
            hsl2rgb(h, m1, m2),
            hsl2rgb(h + 240 if h < 120 else h - 120, m1, m2),
            self.opacity,
        )

    def clamp(self) -> HSL:
        """
        Returns a new HSL color where the h channel is clamped to the range [0, 360),
        and the s, l, and opacity channels are clamped to the range [0, 1].

        Returns
        -------
        HSL
            Clamped color
        """
        return HSL(clamph(self.h), clampt(self.s), clampt(self.l), clampa(self.opacity))

    def displayable(self) -> bool:
        """
        Returns :code:`True` if and only if the color is displayable on standard hardware.
        For example, this returns false for an HSL color if any channel value are in the
        range [0, 1] or if the opacity is not in the range [0, 1].

        Returns
        -------
        bool
            Is displayable
        """
        return (
            (0 <= self.s <= 1 or math.isnan(self.s))
            and (0 <= self.l <= 1)
            and (0 <= self.opacity <= 1)
        )

    def format_hsl(self) -> str:
        a = clampa(self.opacity)
        h = str(clamph(self.h)).removesuffix(".0")
        s = str(clampt(self.s) * 100).removesuffix(".0")
        l = str(clampt(self.l) * 100).removesuffix(".0")
        return f"{'hsl(' if a == 1 else 'hsla('}{h}, {s}%, {l}%{')' if a == 1 else f', {a})'}"

    def __repr__(self) -> str:
        return f"HSL(h={self.h}, s={self.s}, l={self.l}, opacity={self.opacity})"


@overload
def rgb(specifier: str) -> RGB: ...


@overload
def rgb(r: int, g: int, b: int) -> RGB: ...


@overload
def rgb(r: int, g: int, b: int, opacity: float) -> RGB: ...


def rgb(*args):
    """
    Builds a new RGB color.

    Parameters
    ----------
    specifier : str
        String which represents a color
    r : int
        Red value between 0 and 255
    g : int
        Green value between 0 and 255
    b : int
        Blue value between 0 and 255
    opacity : float
        Opacity value between 0 and 1

    Returns
    -------
    RGB
        RGB object

    Examples
    --------

    >>> d3.rgb("#2e65ffff")
    RGB(r=46, g=101, b=255, opacity=1.0)
    >>> d3.rgb(128, 250, 102, 0.2)
    RGB(r=128, g=250, b=102, opacity=0.2)
    """
    if len(args) == 1:
        return rgb_convert(args[0])
    elif len(args) == 3:
        r, g, b = args
        opacity = 1
        return RGB(r, g, b, opacity)
    elif len(args) == 4:
        r, g, b, opacity = args
        return RGB(r, g, b, opacity)


@overload
def hsl(specifier: str) -> HSL: ...


@overload
def hsl(h: float, s: float, l: float) -> HSL: ...


@overload
def hsl(h: float, s: float, l: float, opacity: float) -> HSL: ...


def hsl(*args):
    """
    Build a new HSL color.

    Parameters
    ----------
    specifier : str
        String which represents a color
    h : float
        Hue channel value between 0 and 360
    s : float
        Saturation channel value between 0 and 1
    l : float
        Lightness channel value between 0 and 1
    opacity : float
        Opacity value between 0 and 1

    Returns
    -------
    HSL
        HSL object

    Examples
    --------

    >>> d3.hsl("#2e65ffff")
    HSL(h=224.21052631578948, s=1.0, l=0.5901960784313726, opacity=1.0)
    >>> d3.hsl(210.4, 0.9, 0.7, 0.8)
    HSL(h=210.4, s=0.9, l=0.7, opacity=0.8)
    """
    if len(args) == 1:
        return hsl_convert(args[0])
    elif len(args) == 3:
        h, l, s = args
        opacity = 1
        return HSL(h, l, s, opacity)
    elif len(args) == 4:
        h, l, s, opacity = args
        return HSL(h, l, s, opacity)



from __future__ import annotations

# Since Javascript accepts weird stuff, I had to make the same programming
# "errors" / "mistakes" in order to keep as simple as possible the code. These
# "mistakes" are adding attributes to array object. For instance, in
# Javascript, you can do
# ```js
# var myarray = [1, 2, 3];
# myarray.data = 42
# ```
# In Python, this would raise an error. However, in this specific algorihm,
# lists have a fixed length of 4. I have used this assumption and I dynamically
# add a fifth element (of type `dict`) in `quad` objects in order to *mimic*
# the same behavior.
from collections.abc import Callable
from math import nan, sqrt

from ..array import argpass
from ..quadtree import quadtree
from ..types import SimulationNode, SimulationNodeFunction
from .constant import constant
from .jiggle import jiggle


def x(d: dict[str, float]) -> float:
    return d["x"] + d["vx"]


def y(d: dict[str, float]) -> float:
    return d["y"] + d["vy"]


def quadr(quad: dict | list) -> float:
    return quad["r"] if isinstance(quad, dict) else quad[4]["r"]


class Apply:
    def __init__(self, strength: float, random: Callable[[None], float]):
        self._random = random
        self._strength = strength
        self._node = None
        self._xi = nan
        self._yi = nan
        self._ri = nan
        self._ri2 = nan

    def __call__(
        self, quad: dict | list[dict], x0: float, y0: float, x1: float, y1: float
    ):
        rj = quadr(quad)
        r = self._ri + rj
        if isinstance(quad, dict) and quad["data"]:
            data = quad["data"]
            if data["index"] > self._node["index"]:
                x = self._xi - data["x"] - data["vx"]
                y = self._yi - data["y"] - data["vy"]
                length = x * x + y * y
                if length < r * r:
                    if x == 0:
                        x = jiggle(self._random)
                        length += x * x
                    if y == 0:
                        y = jiggle(self._random)
                        length += y * y

                    length = sqrt(length)
                    length = (r - length) / length * self._strength
                    x *= length
                    y *= length
                    rj *= rj
                    r = rj / (self._ri2 + rj)
                    self._node["vx"] += x * r
                    self._node["vy"] += y * r

                    r = 1 - r
                    data["vx"] -= x * r
                    data["vy"] -= y * r
            return
        return (
            x0 > self._xi + r
            or x1 < self._xi - r
            or y0 > self._yi + r
            or y1 < self._yi - r
        )

    def update(self, node: dict, xi: float, yi: float, ri: float, ri2: float):
        self._node = node
        self._xi = xi
        self._yi = yi
        self._ri = ri
        self._ri2 = ri2


class ForceCollide:
    def __init__(self, radius: SimulationNodeFunction[float]):
        self._radius = radius
        self._nodes = None
        self._radii = None
        self._random = None
        self._strength = 1
        self._iterations = 1

    def __call__(self, alpha: float | None = None):
        apply = Apply(self._strength, self._random)
        for k in range(self._iterations):
            tree = quadtree(self._nodes, x, y).visit_after(self._prepare)
            for node in self._nodes:
                ri = self._radii[node["index"]]
                ri2 = ri * ri
                xi = node["x"] + node["vx"]
                yi = node["y"] + node["vy"]
                apply.update(node, xi, yi, ri, ri2)
                tree.visit(apply)

    def _prepare(self, quad: dict | list[dict]):
        if isinstance(quad, dict) and quad["data"]:
            r = quad["r"] = self._radii[quad["data"]["index"]]
            return r
        if len(quad) == 4:
            quad.append({"r": 0})
        else:  # it assumes there is a fifth element of type `dict`
            quad[4]["r"] = 0
        for i in range(4):
            if quad[i] and quadr(quad[i]) > quad[4]["r"]:
                quad[4]["r"] = quadr(quad[i])

    def _initialize(self):
        if self._nodes is None:
            return
        self._radii = [None] * len(self._nodes)
        for i, node in enumerate(self._nodes):
            self._radii[node["index"]] = self._radius(node, i, self._nodes)

    def initialize(self, nodes: list[SimulationNode], random: Callable[[None], float]):
        self._nodes = nodes
        self._random = random
        self._initialize()

    def set_iterations(self, iterations: int) -> ForceCollide:
        """
        Sets the number of iterations per application to the specified number
        and returns this force.

        Increasing the number of iterations greatly increases the rigidity of
        the constraint and avoids partial overlap of nodes, but also increases
        the runtime cost to evaluate the force.

        Parameters
        ----------
        iterations : int
            Iterations value

        Returns
        -------
        ForceCollide
            Itself
        """
        self._iterations = iterations
        return self

    def set_stength(self, strength: float) -> ForceCollide:
        """
        Sets the force strength to the specified number in the range
        :math:`[0,1]` and returns this force.

        Overlapping nodes are resolved through iterative relaxation. For each
        node, the other nodes that are anticipated to overlap at the next tick
        (using the anticipated positions :math:`(x + v_x,y + v_y)`) are
        determined; the node's velocity is then modified to push the node out
        of each overlapping node. The change in velocity is dampened by the
        force's strength such that the resolution of simultaneous overlaps can
        be blended together to find a stable solution.

        Parameters
        ----------
        strength : float
            Strength value in the range :math:`[0, 1]`

        Returns
        -------
        ForceCollide
            Itself
        """
        self._strength = strength
        return self

    def set_radius(self, radius: SimulationNodeFunction[float] | float) -> ForceCollide:
        """
        Sets the radius accessor to the specified number or function,
        re-evaluates the radius accessor for each node, and returns this force.

        The radius accessor is invoked for each node in the simulation, being
        passed the node and its zero-based index. The resulting number is then
        stored internally, such that the radius of each node is only recomputed
        when the force is initialized or when this method is called with a new
        radius, and not on every application of the force.

        Parameters
        ----------
        radius : SimulationNodeFunction[float] | float
            Radius function or constant value. If it is a function, it takes
            the following arguments:

            * **node** (:code:`SimulationNode`) - the node element
            * **i** (:code:`int`) - the index of the node
            * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

            It returns a radius value (:code:`float`).

        Returns
        -------
        ForceCollide
            Itself
        """
        if callable(radius):
            self._radius = argpass(radius)
        else:
            self._radius = constant(radius)
        self._initialize()
        return self

    def get_iterations(self) -> int:
        return self._iterations

    def get_strength(self) -> float:
        return self._strength

    def get_radius(self) -> SimulationNodeFunction[float]:
        return self._radius


def force_collide(
    radius: SimulationNodeFunction[float] | float | None = None,
) -> ForceCollide:
    """
    The collide force treats nodes as circles with a given radius, rather than
    points, and prevents nodes from overlapping. More formally, two nodes *a*
    and *b* are separated so that the distance between :code:`a` and :code:`b`
    is at least :code:`radius(a) + radius(b)`. To reduce jitter, this is by
    default a "soft" constraint with a configurable strength and iteration
    count.

    Parameters
    ----------
    radius : SimulationNodeFunction[float] | float | None
        Radius function or constant value. If it is a function, it takes
        the following arguments:

        * **node** (:code:`SimulationNode`) - the node element
        * **i** (:code:`int`) - the index of the node
        * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

        It returns a radius value (:code:`float`). If :code:`radius` is not
        specified, the default value is a constant function which returns
        :code:`1`.

    Returns
    -------
    ForceCollide
        Force object
    """
    if not callable(radius):
        radius = constant(1 if radius is None else radius)
    else:
        radius = argpass(radius)
    return ForceCollide(radius)



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import Generic

from ..array import argpass
from ..selection.selection import Selection
from ..types import Accessor, Number, T
from .constant import constant
from .curves import Curve, curve_linear
from .path import WithPath
from .point import x as point_x
from .point import y as point_y


class Line(Generic[T], WithPath):
    """
    The line generator produces a spline or polyline as in a line chart.
    Lines also appear in many other visualization types, such as the links
    in hierarchical edge bundling.

    Parameters
    ----------
    x : Accessor[T, float] | None
        x accessor function for data points
    y : Accessor[T, float] | None
        y accessor function for data points

    Returns
    -------
    Line
        New line generator
    """

    def __init__(
        self, x: Accessor[T, float] | None = None, y: Accessor[T, float] | None = None
    ):
        super().__init__()
        self._defined = constant(True)
        self._context = None
        self._curve = curve_linear
        self._output = None

        if x is None:
            self._x = point_x
        elif callable(x):
            self._x = x
        else:
            self._x = constant(x)

        if y is None:
            self._y = point_y
        elif callable(y):
            self._y = y
        else:
            self._y = constant(y)

        self._x = argpass(self._x)
        self._y = argpass(self._y)

    def __call__(self, data: Iterable[T]) -> str | None:
        """
        Generate a line for the given list of data

        Parameters
        ----------
        data : Iterable[T]
            Data values

        Returns
        -------
        str | None
            Generated line if the line is not associated to a context

        Examples
        --------

        >>> svg.append("path").attr("d", line(data)).attr("stroke", "currentColor")
        """
        data = list(data)
        n = len(data)
        if n == 0:
            return None
        defined0 = False

        buffer = None
        if self._context is None:
            buffer = self._path()
            self._output = self._curve(buffer)

        for i in range(n + 1):
            d = data[i] if i < n else None
            if not (i < n and self._defined(d, i, data) == defined0):
                defined0 = not defined0
                if defined0:
                    self._output.line_start()
                else:
                    self._output.line_end()
            if defined0:
                self._output.point(self._x(d, i, data), self._y(d, i, data))

        if buffer:
            self._output = None
            return str(buffer) or None

    def x(self, x: Accessor[T, float] | Number) -> Line:
        """
        Sets x accessor function

        Parameters
        ----------
        x : Accessor[T, float] | Number
            x accessor function

        Returns
        -------
        Line
            Itself
        """
        if callable(x):
            self._x = x
        else:
            self._x = constant(x)
        self._x = argpass(self._x)
        return self

    def y(self, y: Accessor[T, float] | Number) -> Line:
        """
        Sets y accessor function

        Parameters
        ----------
        y : Accessor[T, float] | Number
            y accessor function

        Returns
        -------
        Line
            Itself
        """
        if callable(y):
            self._y = y
        else:
            self._y = constant(y)
        self._y = argpass(self._y)
        return self

    def set_defined(self, defined: Accessor[T, bool] | Number) -> Line:
        """
        Sets defined accessor

        When a line is generated, the defined accessor will be invoked
        for each element in the input data array, being passed the element
        :code:`d`, the index :code:`i`, and the array :code:`data` as three
        arguments. If the given element is defined (i.e., if the defined accessor
        returns a truthy value for this element), the x and y accessors will
        subsequently be evaluated and the point will be added to the current
        line segment. Otherwise, the element will be skipped, the current line
        segment will be ended, and a new line segment will be generated for the
        next defined point.

        Parameters
        ----------
        defined : Accessor[T, bool] | Number
            defined accessor function

        Returns
        -------
        Line
            Itself
        """
        if defined is None:
            self._defined = defined
        elif callable(defined):
            self._defined = defined
        else:
            self._defined = constant(bool(defined))
        return self

    def set_curve(self, curve: Callable[[Selection], Curve] | None = None) -> Line:
        """
        Sets curve factory.

        Parameters
        ----------
        curve : Callable[[Selection], Curve] | None
            Curve factory function

        Returns
        -------
        Line
            Itself
        """
        self._curve = curve
        if self._context is not None:
            self._output = self._curve(self._context)
        return self

    def set_context(self, context: Selection | None = None) -> Line:
        """
        Sets the context.

        Parameters
        ----------
        context : Selection | None
            Selection

        Returns
        -------
        Line
            Itself
        """
        if context is None:
            self._context = None
            self._output = None
        else:
            self._context = context
            self._output = self._curve(self._context)
        return self

    def get_x(self) -> Accessor[T, float]:
        return self._x

    def get_y(self) -> Accessor[T, float]:
        return self._y

    def get_defined(self) -> Accessor[T, bool]:
        return self._defined

    def get_curve(self) -> Callable[[Selection], Curve]:
        return self._curve

    def get_context(self) -> Selection:
        return self._context



from ..hierarchy import Node


def slice(parent: Node, x0: float, y0: float, x1: float, y1: float):
    """
    Divides the rectangular area specified by :math:`x_0:, :math:`y_0`,
    :math:`x_1`, :math:`y_1` vertically according the value of each of the
    specified node's children. The children are positioned in order, starting
    with the top edge (:math:`y_0`) of the given rectangle. If the sum of the
    children’s values is less than the specified node’s value (i.e., if the
    specified node has a non-zero internal value), the remaining empty space
    will be positioned on the bottom edge (:math:`y_1`) of the given rectangle.

    Parameters
    ----------
    parent : Node
        Parent node
    x0 : float
        X-coordinate top edge
    y0 : float
        Y-coordinate top edge
    x1 : float
        X-coordinate bottom edge
    y1 : float
        Y-coordinate bottom edge
    """
    nodes = parent.children
    k = 0 if parent.value is None or parent.value == 0 else (y1 - y0) / parent.value

    for node in nodes:
        node.x0 = x0
        node.x1 = x1
        node.y0 = y0
        y0 += node.value * k
        node.y1 = y0



import re

from lxml import etree

from ..array import argpass
from ..types import Accessor, EtreeFunction, T

CLASS_PATTERN = re.compile(r"^|\s+")


def class_array(string: str) -> list[str]:
    """
    Splits a class string into several class names.

    Parameters
    ----------
    string : str
        Class string

    Returns
    -------
    list[str]
        Class names
    """
    return CLASS_PATTERN.split(string.strip())


def classed_add(node: etree.Element, names: list[str]):
    """
    Adds the list of :code:`names` into class property of the specified node.

    Parameters
    ----------
    node : etree.Element
        Node to update
    names : list[str]
        Class names
    """
    class_names = []
    if class_string := node.get("class"):
        class_names = class_string.split(" ")
    for name in names:
        if name not in class_names:
            class_names.append(name)
    node.set("class", " ".join(class_names).strip())


def classed_remove(node: etree.Element, names: list[str]):
    """
    Removes the list of :code:`names` into class property of the specified
    node.

    Parameters
    ----------
    node : etree.Element
        Node to update
    names : list[str]
        Class names
    """
    class_names = []
    if class_string := node.get("class"):
        class_names = class_string.split(" ")
    length = len(class_names)
    for name in names:
        # Not clean but faster than checking if `class_names` contain `name`
        try:
            class_names.remove(name)
        except ValueError:
            pass
    if len(class_names) != length:
        node.set("class", " ".join(class_names).strip())


def classed_constant(class_names: str, value: bool) -> EtreeFunction[T, None]:
    """
    Returns a function which adds or removes specified classes of a specified
    node given a condition :code:`value`.

    Parameters
    ----------
    class_names : str
        Class name
    value : bool
        Constant condition for adding or removing class name

    Returns
    -------
    EtreeFunction[T, None]
        Function for adding or removing class name
    """
    names = class_array(class_names)
    classed = classed_add if value else classed_remove

    def callback(node: etree.Element, data: T, i: int, group: list[etree.Element]):
        classed(node, names)

    return callback


def classed_function(
    class_names: str,
    value: Accessor[T, str | bool],
) -> EtreeFunction[T, None]:
    """
    Returns a function which adds or removes specified classes of a specified
    node given a condition :code:`value`.

    Parameters
    ----------
    class_names : str
        Class name
    value : Accessor[T, str | bool]
        Condition function which returns a boolean indicating if the class name
        must be added or removed from the node.

    Returns
    -------
    EtreeFunction[T, None]
        Function for adding or removing class name
    """
    names = class_array(class_names)
    value = argpass(value)

    def callback(node: etree.Element, data: T, i: int, group: list[etree.Element]):
        if value(data, i, group):
            classed_add(node, names)
        else:
            classed_remove(node, names)

    return callback



from typing import Any

from lxml import etree

from ..array import argpass
from ..types import Accessor, EtreeFunction, T


def text_constant(value: Any) -> EtreeFunction[T, None]:
    """
    Returns a function which adds text value to nodes given a constant
    value.

    Parameters
    ----------
    value : Any
        Constant text value

    Returns
    -------
    EtreeFunction[T, None]
        Function which adds a text value to nodes
    """
    value = str(value)

    def callback(node: etree.Element, data: T, i: int, group: list[etree.Element]):
        node.text = value

    return callback


def text_function(value: Accessor[T, Any]) -> EtreeFunction[T, None]:
    """
    Returns a function which adds text value to nodes given a constant
    value.

    Parameters
    ----------
    value : Accessor[T, str]
        Accessor function

    Returns
    -------
    EtreeFunction[T, None]
        Function which adds a text value to nodes
    """
    value = argpass(value)

    def callback(node: etree.Element, data: T, i: int, group: list[etree.Element]):
        node.text = str(value(data, i, group))

    return callback



from __future__ import annotations

from collections.abc import Callable
from datetime import datetime
from typing import overload

from ..time import (
    time_day,
    time_hour,
    time_minute,
    time_month,
    time_second,
    time_tick_interval,
    time_ticks,
    time_week,
    time_year,
)
from ..time_format import time_format
from ..types import Formatter, T
from .continuous import Transformer, copy, identity
from .init import init_range
from .nice import nice


class ScaleTime(Transformer[datetime]):
    """
    Time scales are a variant of linear scales that have a temporal domain:
    domain values are coerced to dates rather than numbers, and invert
    likewise returns a date. Time scales implement ticks based on calendar
    intervals, taking the pain out of generating axes for temporal domains.

    Parameters
    ----------
    ticks : Callable
        Ticks function
    tick_interval : Callable
        Tick interval function
    year : Callable
        Year time function
    month : Callable
        Month time function
    week : Callable
        Week time function
    day : Callable
        Day time function
    hour : Callable
        Hour time function
    minute : Callable
        Minute time function
    second : Callable
        Second time function
    """

    def __init__(
        self,
        ticks: Callable,
        tick_interval: Callable,
        year: Callable,
        month: Callable,
        week: Callable,
        day: Callable,
        hour: Callable,
        minute: Callable,
        second: Callable,
    ):
        super().__init__(identity, identity)

        self._ticks = ticks
        self._tick_interval = tick_interval
        self._year = year
        self._month = month
        self._week = week
        self._day = day
        self._hour = hour
        self._minute = minute
        self._second = second
        self._format = time_format

        self._format_millisecond = time_format(".%L")
        self._format_second = time_format(":%S")
        self._format_minute = time_format("%I:%M")
        self._format_hour = time_format("%I %p")
        self._format_day = time_format("%a %d")
        self._format_week = time_format("%b %d")
        self._format_month = time_format("%B")
        self._format_year = time_format("%Y")

    def _tick_format(self, date: datetime | None = None) -> Formatter[str]:
        """
        Private method which generates a formatter function given the "magnitude" of
        the date input.

        Parameters
        ----------
        date : datetime | None
            Date input

        Returns
        -------
        Formatter[str]
            Formatter function which returns a string
        """
        if self._second(date) < date:
            return self._format_millisecond(date)
        elif self._minute(date) < date:
            return self._format_second(date)
        elif self._hour(date) < date:
            return self._format_minute(date)
        elif self._day(date) < date:
            return self._format_hour(date)
        elif self._month(date) < date:
            if self._week(date) < date:
                return self._format_day(date)
            else:
                return self._format_week(date)
        elif self._year(date) < date:
            return self._format_month(date)
        else:
            return self._format_year(date)

    def invert(self, y: T) -> datetime:
        """
        Given a value from the range, returns the corresponding value
        from the domain. Inversion is useful for interaction, say to
        determine the data value corresponding to the position of the mouse.

        Parameters
        ----------
        y : T
            Input value

        Returns
        -------
        datetime
            Corresponding value from the domain
        """
        return datetime.fromtimestamp(super().invert(y))

    def ticks(self, count: int | None = None) -> list[datetime]:
        """
        Returns representative dates from the scale’s domain.

        Parameters
        ----------
        count : int | None
            Count may be specified to affect how many ticks
            are generated. If count is not specified, it
            defaults to 10.

        Returns
        -------
        list[datetime]
            The returned tick values are uniformly-spaced (mostly),
            have sensible values (such as every day at midnight),
            and are guaranteed to be within the extent of the domain.
            Ticks are often used to display reference lines, or tick
            marks, in conjunction with the visualized data.
        """
        d = self.get_domain()
        return self._ticks(d[0], d[-1], count if count is not None else 10)

    def tick_format(self, _: int = 0, specifier: str | None = None) -> Formatter[str]:
        """
        Returns a formatter function suitable for displaying a tick value,
        automatically computing the appropriate precision based on the fixed interval
        between tick values. The specified count should have the same value as the
        count that is used to generate the tick values.

        Parameters
        ----------
        _ : int
            Count parameter, not used in this function, only exists
            for generic signature method
        specifier : str | None
            Specifier

        Returns
        -------
        Formatter[str]
            Tick format function which returns a string
        """
        return self._tick_format if specifier is None else self._format(specifier)

    def nice(self, interval: Callable | list[datetime] | None = None) -> ScaleTime:
        """
        This method typically modifies the scale’s domain, and may only extend the
        bounds to the nearest round value.

        Parameters
        ----------
        interval : Callable | list[datetime] | None
            Argument which allows greater control over the step size used
            to extend the bounds, guaranteeing that the returned ticks
            will exactly cover the domain.

        Returns
        -------
        ScaleTime
            Itself
        """
        d = self.get_domain()
        if not interval or not hasattr(interval, "range"):
            interval = self._tick_interval(
                d[0], d[-1], interval if interval is not None else 10
            )
        return self.set_domain(nice(d, interval)) if interval else self

    def copy(self):
        return copy(
            self,
            ScaleTime(
                self._ticks,
                self._tick_interval,
                self._year,
                self._month,
                self._week,
                self._day,
                self._hour,
                self._minute,
                self._second,
            ),
        )


@overload
def scale_time() -> ScaleTime: ...


@overload
def scale_time(range_vals: list[T]) -> ScaleTime: ...


@overload
def scale_time(domain: list[datetime], range_vals: list[T]) -> ScaleTime: ...


def scale_time(*args) -> ScaleTime:
    """
    Builds a new time scale with the specified domain and range,
    the default interpolator and clamping disabled

    Parameters
    ----------
    domain : list[datetime]
        Array of datetime
    range_vals : list[T]
        Array of values


    Returns
    -------
    ScaleTime
        Scale object

    Examples
    --------

    >>> from datetime import datetime
    >>> scale = d3.scale_time([datetime(2000, 1, 1), datetime(2000, 1, 2)], [0, 960])
    >>> for x in range(24 + 1):
    ...     x = datetime(2000, 1, 1) + x * hour
    ...     print(x, scale(x))
    ...
    ...
    2000-01-01 00:00:00 0.0
    2000-01-01 01:00:00 40.0
    2000-01-01 02:00:00 80.0
    2000-01-01 03:00:00 120.0
    2000-01-01 04:00:00 160.0
    2000-01-01 05:00:00 200.0
    2000-01-01 06:00:00 240.0
    2000-01-01 07:00:00 280.0
    2000-01-01 08:00:00 320.0
    2000-01-01 09:00:00 360.0
    2000-01-01 10:00:00 400.0
    2000-01-01 11:00:00 440.0
    2000-01-01 12:00:00 480.0
    2000-01-01 13:00:00 520.0
    2000-01-01 14:00:00 560.0
    2000-01-01 15:00:00 600.0
    2000-01-01 16:00:00 640.0
    2000-01-01 17:00:00 680.0
    2000-01-01 18:00:00 720.0
    2000-01-01 19:00:00 760.0
    2000-01-01 20:00:00 800.0
    2000-01-01 21:00:00 840.0
    2000-01-01 22:00:00 880.0
    2000-01-01 23:00:00 920.0
    2000-01-02 00:00:00 960.0
    """
    scaler = ScaleTime(
        time_ticks,
        time_tick_interval,
        time_year,
        time_month,
        time_week,
        time_day,
        time_hour,
        time_minute,
        time_second,
    ).set_domain([datetime(2000, 1, 1), datetime(2000, 1, 2)])
    if len(args) == 1:
        return init_range(scaler, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scaler, domain=domain, range_vals=range_vals)
    return init_range(scaler)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "fc8d59ffffbf91cf60",
    "d7191cfdae61a6d96a1a9641",
    "d7191cfdae61ffffbfa6d96a1a9641",
    "d73027fc8d59fee08bd9ef8b91cf601a9850",
    "d73027fc8d59fee08bffffbfd9ef8b91cf601a9850",
    "d73027f46d43fdae61fee08bd9ef8ba6d96a66bd631a9850",
    "d73027f46d43fdae61fee08bffffbfd9ef8ba6d96a66bd631a9850",
    "a50026d73027f46d43fdae61fee08bd9ef8ba6d96a66bd631a9850006837",
    "a50026d73027f46d43fdae61fee08bffffbfd9ef8ba6d96a66bd631a9850006837",
]
SCHEME_RDYLGN = list(map(colors, scheme))
interpolate_rdylgn = ramp(SCHEME_RDYLGN)



from ...types import GeoJSON, Point2D, Vec2D
from ..common import PolygonStream, Projection
from .albers import geo_albers
from .conic_equal_area import geo_conic_equal_area
from .fit import fit_extent, fit_height, fit_size, fit_width

EPSILON = 1e-6


class Multiplex:
    def __init__(self, streams: list[PolygonStream]):
        self._streams = streams

    def point(self, x: float, y: float):
        for stream in self._streams:
            stream.point(x, y)

    def line_start(self):
        for stream in self._streams:
            stream.line_start()

    def line_end(self):
        for stream in self._streams:
            stream.line_end()

    def polygon_start(self):
        for stream in self._streams:
            stream.polygon_start()

    def polygon_end(self):
        for stream in self._streams:
            stream.polygon_end()

    def sphere(self):
        for stream in self._streams:
            stream.sphere()


class PointStream:
    def __init__(self):
        self._point = None

    def point(self, x: float, y: float):
        self._point = [x, y]

    def __str__(self) -> str:
        return f"AlbertUsaPointStream({self._point})"


class AlbersUsaProjection(Projection):
    def __init__(self):
        self._cache = None
        self._cache_stream = None
        self._lower48 = geo_albers()
        self._lower48_point = None
        self._alaska = (
            geo_conic_equal_area()
            .rotate([154, 0])
            .set_center([-2, 58.5])
            .parallels([55, 65])
        )
        self._alaska_point = None
        self._hawaii = (
            geo_conic_equal_area()
            .rotate([157, 0])
            .set_center([-3, 19.9])
            .parallels([8, 18])
        )
        self._hawaii_point = None
        self._point_stream = PointStream()

    def __call__(self, coordinates: Point2D) -> Point2D:
        """
        Returns a new array :code:`[x, y]` (typically in pixels) representing
        the projected point of the given point. The point must be specified as
        a two-element array :code:`[longitude, latitude]` in degrees. May
        return :code:`None` if the specified point has no defined projected position,
        such as when the point is outside the clipping bounds of the
        projection.

        Parameters
        ----------
        point : Point2D
            Point :code:`[longitude, latitude]`

        Returns
        -------
        Point2D
            New projected point :code:`[x, y]`
        """
        x = coordinates[0]
        y = coordinates[1]
        self._point_stream._point = None
        self._lower48_point.point(x, y)
        if point := self._point_stream._point:
            return point
        self._alaska_point.point(x, y)
        if point := self._point_stream._point:
            return point
        self._hawaii_point.point(x, y)
        if point := self._point_stream._point:
            return point

    def invert(self, coordinates: Point2D) -> Point2D:
        """
        Returns a new array :code:`[longitude, latitude]` in degrees
        representing the unprojected point of the given projected point. The
        point must be specified as a two-element array :code:`[x, y]`
        (typically in pixels). May return null if the specified point has no
        defined projected position, such as when the point is outside the
        clipping bounds of the projection.

        This method is only defined on invertible projections.

        Parameters
        ----------
        point : Point2D
            2D point where coordinates are in pixels

        Returns
        -------
        Point2D
            2D point where coordinates are in degrees
        """
        k = self._lower48.get_scale()
        t = self._lower48.get_translation()
        x = (coordinates[0] - t[0]) / k
        y = (coordinates[1] - t[1]) / k
        if 0.120 <= y < 0.234 and -0.425 <= x < -0.214:
            return self._alaska.invert(coordinates)
        if 0.166 <= y < 0.234 and -0.214 <= x < -0.115:
            return self._hawaii.invert(coordinates)
        else:
            return self._lower48.invert(coordinates)

    def stream(self, stream: PointStream):
        """
        Returns a projection stream for the specified output stream.

        Parameters
        ----------
        stream : PolygonStream
            Stream object

        Returns
        -------
        PolygonStream
            Modified stream object
        """
        if self._cache is not None and self._cache_stream == stream:
            return self._cache
        else:
            self._cache_stream = stream
            self._cache = Multiplex(
                [
                    self._lower48.stream(stream),
                    self._alaska.stream(stream),
                    self._hawaii.stream(stream),
                ]
            )
            return self._cache

    def set_precision(self, precision: float) -> Projection:
        """
        If precision is specified, sets the threshold for the projection's
        adaptive resampling to the specified value in pixels and returns the
        projection. This value corresponds to the Douglas–Peucker distance.

        Parameters
        ----------
        precision : float
            Precision value

        Returns
        -------
        Projection
            Itself
        """
        self._lower48.set_precision(precision)
        self._alaska.set_precision(precision)
        self._hawaii.set_precision(precision)
        return self.reset()

    def scale(self, k: float) -> Projection:
        """
        If scale is specified, sets the projection's scale factor to the
        specified value and returns the projection. The scale factor
        corresponds linearly to the distance between projected points; however,
        absolute scale factors are not equivalent across projections.

        Parameters
        ----------
        k : float
            Scale factor

        Returns
        -------
        Projection
            Itself
        """
        self._lower48.scale(k)
        self._alaska.scale(k * 0.35)
        self._hawaii.scale(k)
        return self.translate(self._lower48.get_translation())

    def translate(self, translation: Vec2D) -> Projection:
        """
        If translate is specified, sets the projection's translation offset to
        the specified two-element array :code:`[tx, ty]` and returns the
        projection. The translation offset determines the pixel coordinates of
        the projection's center.

        Parameters
        ----------
        translation : Vec2D
            Translation 2D vector

        Returns
        -------
        Projection
            Itself
        """
        k = self._lower48.get_scale()
        x = translation[0]
        y = translation[1]

        self._lower48_point = (
            self._lower48.translate(translation)
            .set_clip_extent(
                [[x - 0.455 * k, y - 0.238 * k], [x + 0.455 * k, y + 0.238 * k]]
            )
            .stream(self._point_stream)
        )

        self._alaska_point = (
            self._alaska.translate([x - 0.307 * k, y + 0.201 * k])
            .set_clip_extent(
                [
                    [x - 0.425 * k + EPSILON, y + 0.120 * k + EPSILON],
                    [x - 0.214 * k - EPSILON, y + 0.234 * k - EPSILON],
                ]
            )
            .stream(self._point_stream)
        )

        self._hawaii_point = (
            self._hawaii.translate([x - 0.205 * k, y + 0.212 * k])
            .set_clip_extent(
                [
                    [x - 0.214 * k + EPSILON, y + 0.166 * k + EPSILON],
                    [x - 0.115 * k - EPSILON, y + 0.234 * k - EPSILON],
                ]
            )
            .stream(self._point_stream)
        )

        return self.reset()

    def fit_extent(self, extent: tuple[Point2D, Point2D], obj: GeoJSON) -> Projection:
        """
        Sets the projection's scale and translate to fit the specified GeoJSON
        object in the center of the given extent. The extent is specified as an
        array :math:`[[x_0, y_0], [x_1, y_1]]`, where :math:`x_0` is the left
        side of the bounding box, :math:`y_0` is the top, :math:`x_1`₁ is the
        right and :math:`y_1` is the bottom. Returns the projection.

        Any clip extent is ignored when determining the new scale and
        translate. The precision used to compute the bounding box of the given
        object is computed at an effective scale of 150.

        Parameters
        ----------
        extent : tuple[Point2D, Point2D]
            Extent values
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        return fit_extent(self, extent, obj)

    def fit_size(self, size: Point2D, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_extent where the top-left
        corner of the extent is [0, 0].

        Parameters
        ----------
        size : Point2D
            Size values
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        return fit_size(self, size, obj)

    def fit_width(self, width: float, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_size where the height is
        automatically chosen from the aspect ratio of object and the given
        constraint on width.

        Parameters
        ----------
        width : float
            Width value
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        return fit_width(self, width, obj)

    def fit_height(self, height: float, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_size where the width is
        automatically chosen from the aspect ratio of object and the given
        constraint on height.

        Parameters
        ----------
        height : float
            Height value
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        return fit_height(self, height, obj)

    def reset(self):
        self._cache = None
        self._cache_stream = None
        return self

    def get_precision(self) -> float:
        return self._lower48.get_precision()

    def get_scale(self) -> float:
        return self._lower48.get_scale()

    def get_translation(self) -> Vec2D:
        return self._lower48.get_translation()


def geo_albers_usa() -> Projection:
    """
    This is a U.S.-centric composite projection of three geoConicEqualArea
    projections: :func:`d3.geo_albers <geo_albers>` is used for the lower
    forty-eight states, and separate conic equal-area projections are used for
    Alaska and Hawaii. The scale for Alaska is diminished: it is projected at
    :math:`0.35 \\times` its true relative area.

    The constituent projections have fixed clip, center and rotation, and thus
    this projection does not support :func:`Projection.set_center
    <detroit.geo.common.Projection.set_center>`, :func:`Projection.rotate
    <detroit.geo.common.Projection.rotate>`, :func:`Projection.set_clip_angle
    <detroit.geo.common.Projection.set_clip_angle>` or
    :func:`Projection.set_clip_extent
    <detroit.geo.common.Projection.set_clip_extent>`.

    Returns
    -------
    Projection
        Projection object
    """
    return AlbersUsaProjection().scale(1070)



from ..hierarchy import Node


def binary(parent: Node, x0: float, y0: float, x1: float, y1: float):
    """
    Recursively partitions the specified nodes into an approximately-balanced
    binary tree, choosing horizontal partitioning for wide rectangles and
    vertical partitioning for tall rectangles.

    Parameters
    ----------
    parent : Node
        Parent node
    x0 : float
        X-coordinate rectangular edge
    y0 : float
        Y-coordinate rectangular edge
    x1 : float
        X-coordinate rectangular edge
    y1 : float
        Y-coordinate rectangular edge
    """
    nodes = parent.children
    n = len(nodes)
    sums = [None] * (n + 1)

    sum_value = sums[0] = 0
    for i in range(n):
        sum_value += nodes[i].value
        sums[i + 1] = sum_value

    def partition(i, j, value, x0, y0, x1, y1):
        if i >= j - 1:
            node = nodes[i]
            node.x0 = x0
            node.x1 = x1
            node.y0 = y0
            node.y1 = y1
            return

        value_offset = sums[i]
        value_target = (value * 0.5) + value_offset
        k = i + 1
        hi = j - 1

        while k < hi:
            mid = k + hi >> 1
            if sums[mid] < value_target:
                k = mid + 1
            else:
                hi = mid

        if (value_target - sums[k - 1]) < (sums[k] - value_target) and i + 1 < k:
            k -= 1

        value_left = sums[k] - value_offset
        value_right = value - value_left

        if (x1 - x0) > (y1 - y0):
            xk = (x0 * value_right + x1 * value_left) / value if value else x1
            partition(i, k, value_left, x0, y0, xk, y1)
            partition(k, j, value_right, xk, y0, x1, y1)
        else:
            yk = (y0 * value_right + y1 * value_left) / value if value else y1
            partition(i, k, value_left, x0, y0, x1, yk)
            partition(k, j, value_right, x0, yk, x1, y1)

    partition(0, n, parent.value, x0, y0, x1, y1)



import math

from ..path import Path


class WithPath:
    def __init__(self):
        self._digits = 3

    def digits(self, digits=None):
        if digits is None:
            return self._digits
        d = math.floor(digits)
        if d < 0:
            raise ValueError(f"Invalid digits: {d}")
        self._digits = d
        return self

    def _path(self):
        return Path(self._digits)



from lxml import etree

from .selection import Selection


def select(node: etree.Element) -> Selection:
    """
    Returns a selection object given a node

    Parameters
    ----------
    node : etree.Element
        Node

    Returns
    -------
    Selection
        Selection object
    """
    return Selection([[node]], [node])



from __future__ import annotations

import math
from typing import Callable, Generic, overload

from ..array import tick_increment, ticks
from ..types import Number, T
from .continuous import Transformer, copy
from .init import init_range
from .tick_format import tick_format


class LinearBase:
    def ticks(self, count: int | None = None) -> list[Number]:
        """
        Returns approximately count representative values
        from the scale's domain.

        Parameters
        ----------
        count : int | None
            Count. If specified, the scale may return more or fewer
            values depending on the domain

        Returns
        -------
        list[Number]
            Tick values are uniformly spaced, have human-readable values
            (such as multiples of powers of 10), and are guaranteed to be
            within the extent of the domain. Ticks are often used to display
            reference lines, or tick marks, in conjunction with the visualized data.
        """
        d = self.get_domain()
        return ticks(d[0], d[-1], count if count is not None else 10)

    def tick_format(
        self, count: int | None = None, specifier: str | None = None
    ) -> Callable[[Number], str]:
        """
        Returns a number format function suitable for displaying
        a tick value, automatically computing the appropriate
        precision based on the fixed interval between tick values.
        The specified count should have the same value as the count
        that is used to generate the tick values.

        Parameters
        ----------
        count : int | None
            Count. It should have the same value as the count
            that is used to generate the tick values.
        specifier : str | None
            Specifier

        Returns
        -------
        Callable[[Number], str]
            Tick format function
        """
        d = self.get_domain()
        return tick_format(d[0], d[-1], count if count is not None else 10, specifier)

    def nice(self, count: int | None = None) -> LinearBase:
        """
        Extends the domain so that it starts and ends on nice round values.

        Parameters
        ----------
        count : int | None
            Count argument allows greater control over the step size
            used to extend the bounds, guaranteeing that the returned
            ticks will exactly cover the domain

        Returns
        -------
        LinearBase
            Itself
        """
        if count is None:
            count = 10

        d = self.get_domain()
        i0 = 0
        i1 = len(d) - 1
        start = d[i0]
        stop = d[i1]
        prestep = None
        step = None
        max_iter = 10

        if stop < start:
            step = start
            start = stop
            stop = step
            step = i0
            i0 = i1
            i1 = step

        while max_iter > 0:
            step = tick_increment(start, stop, count)
            if step == prestep:
                d[i0] = start
                d[i1] = stop
                return self.set_domain(d)
            elif step > 0:
                start = math.floor(start / step) * step
                stop = math.ceil(stop / step) * step
            elif step < 0:
                start = math.ceil(start * step) / step
                stop = math.floor(stop * step) / step
            else:
                break
            prestep = step
            max_iter -= 1

        return self

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


class ScaleLinear(Transformer[T], LinearBase, Generic[T]):
    """
    Linear scales map a continuous, quantitative input domain to a continuous
    output range using a linear transformation (translate and scale). If the
    range is also numeric, the mapping may be inverted. Linear scales are a good
    default choice for continuous quantitative data because they preserve
    proportional differences. Each range value y can be expressed as a function
    of the domain value x: :math:`y = m \\cdot x + b`.

    Parameters
    ----------
    t : Callable[[Number], T]
        Transform function
    u : Callable[[T], Number]
        Untransform function
    """

    def copy(self):
        return copy(self, ScaleLinear())


@overload
def scale_linear() -> ScaleLinear[T]: ...


@overload
def scale_linear(range_vals: list[T]) -> ScaleLinear[T]: ...


@overload
def scale_linear(domain: list[Number], range_vals: list[T]) -> ScaleLinear[T]: ...


def scale_linear(*args):
    """
    Builds a new linear scale with the specified domain
    and range, the default interpolator, and clamping disabled.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    range_vals : list[T]
        Array of values

    Returns
    -------
    ScaleLinear[T]
        Scale object

    Examples
    --------

    >>> scale = d3.scale_linear([0, 100], ["red", "blue"])
    >>> for x in range(11):
    ...     x = x * 10
    ...     print(x, scale(x))
    ...
    ...
    0 rgb(255, 0, 0)
    10 rgb(230, 0, 26)
    20 rgb(204, 0, 51)
    30 rgb(178, 0, 76)
    40 rgb(153, 0, 102)
    50 rgb(128, 0, 128)
    60 rgb(102, 0, 153)
    70 rgb(76, 0, 178)
    80 rgb(51, 0, 204)
    90 rgb(26, 0, 230)
    100 rgb(0, 0, 255)
    """
    scale = ScaleLinear()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from ...types import T
from ..series import Series


def offset_diverging(series: list[Series[T]], order: list[int]):
    """
    Positive values are stacked above zero, negative values are stacked below
    zero, and zero values are stacked at zero.

    Parameters
    ----------
    series : list[Series[T]]
        List of series
    order : list[int]
        List of ordered indices
    """
    n = len(series)
    if n == 0:
        return
    m = len(series[order[0]])
    for j in range(m):
        yp = yn = 0.0
        for i in range(n):
            d = series[order[i]][j]
            dy = d[1] - d[0]
            if dy > 0:
                d[0] = yp
                yp += dy
                d[1] = yp
            elif dy < 0:
                d[1] = yn
                yn += dy
                d[0] = yn
            else:
                d[0] = 0
                d[1] = dy



from collections.abc import Callable

from ..types import T


def interpolate_discrete(range_: list[T]) -> Callable[[float], T]:
    """
    Returns a discrete interpolator for the given array of values.

    Parameters
    ----------
    range_ : list[T]
        List of elements

    Returns
    -------
    Callable[[float], T]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_discrete(["a", "b", "c"])
    >>> interpolator(0)
    'a'
    >>> interpolator(1)
    'c'
    >>> interpolator(0.5)
    'b'
    """
    n = len(range_)
    return lambda t: range_[max(0, min(n - 1, int(t * n)))]
